#! /usr/bin/env python

import numpy as n, aipy as a, pylab as p, random, sys
try: from mpl_toolkits.basemap import Basemap
except(ImportError): from matplotlib.toolkits.basemap import Basemap

NPTS = 100
random.seed(1)

ra,dec = a.map.facet_centers(NPTS, ncrd=2)

slats = dec * a.img.rad2deg
slons = ra * a.img.rad2deg
slons -= 180
slons = n.where(slons < -180, slons +360, slons)

map = Basemap(projection='moll', lat_0=0, lon_0=0)
map.drawmapboundary()
map.drawmeridians(n.arange(0,360,30))
map.drawparallels(n.arange(-90,90,30))
sx, sy = map(slons, slats)
map.plot(sx, sy, 'ko')
for c, (xpt,ypt) in enumerate(zip(sx, sy)):
    p.text(xpt+50000, ypt+50000, str(c))
p.show()



#! /usr/bin/env python
"""
A script for fitting the amplitudes and passbands of antennas given
starting parameters in "a.loc" and a list of sources (from "a.src").

Author: Aaron Parsons
"""

import aipy as a, numpy as n, sys, os, optparse

o = optparse.OptionParser()
o.set_usage('fitmdl.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, ant=True, pol=True, chan=True,
    loc=True, src=True, dec=True)
o.add_option('--aprms', dest='aprms',
    help='Comma delimited list of paramters to fit independently for each antenna.')
o.add_option('--fitants', dest='fitants', default='all',
    help='Comma delimited list of antennas to fit. Default "all".')
o.add_option('--aaprms', dest='aaprms',
    help='Comma delimited list of AntennaArray (not Antenna) parameters to fit.  Unless added specifically in loc file, default AntennaArray has no parameters to fit.')
o.add_option('--shprms', dest='shprms',
    help='Comma delimited list of parameters to fit whose values will be shared for all antennas.')
o.add_option('--sprms', dest='sprms', 
    help='Source=param pairs for fitting source parameters.')
o.add_option('--sim_autos', dest='sim_autos', action='store_true',
    help='Use auto-correlations in fitting.  Default is to use only cross-correlations.')
o.add_option('-n', '--norm', dest='norm', action='store_true',
    help='Normalize residual to data strength.  Puts weaker sources on similar footing as strong sources.')
opts, args = o.parse_args(sys.argv[1:])

# Parse command-line options
uv = a.miriad.UV(args[0])
a.scripting.uv_selector(uv, opts.ant, opts.pol)
aa = a.loc.get_aa(opts.loc, uv['sdf'], uv['sfreq'], uv['nchan'])
chans = a.scripting.parse_chans(opts.chan, uv['nchan'])
aa.select_chans(chans)
cat = a.scripting.parse_srcs(opts.src, force_cat=True)
cat.set_params(a.loc.get_src_prms(opts.loc))
(uvw,t,(i,j)),d = uv.read()
aa.set_jultime(t)
cat.compute(aa)
if opts.fitants.startswith('all'): fitants = range(uv['nants'])
else: fitants = map(int, opts.fitants.split(','))
del(uv)

# Generate list of parameters to fit
if opts.aprms is None: aprms = []
else: aprms = opts.aprms.split(',')
if opts.shprms is None: shprms = []
else: shprms = opts.shprms.split(',')
if opts.aaprms is None: aaprms = []
else: aaprms = opts.aaprms.split(',')
prm_dict = {}
for ant in fitants: prm_dict[ant] = aprms[:]
prm_dict[fitants[0]].extend(shprms)
prm_dict['aa'] = aaprms
print prm_dict
start_prms = aa.get_params(prm_dict)
if opts.sprms != None:
    sprms = [w.split('=') for w in opts.sprms.split(',')]
    sprm_dict = {}
    for src,prm in sprms:
        if sprm_dict.has_key(src): sprm_dict[src].append(prm)
        else: sprm_dict[src] = [prm]
    start_prms.update(cat.get_params(sprm_dict))
else: sprms = []
prm_list, key_list = a.fit.flatten_prms(start_prms)

first_fit = None    # Used to normalize fit values to the starting fit
mfq = cat.get_mfreqs()

def fit_func(prms):
    global first_fit
    prms = a.fit.reconstruct_prms(prms, key_list)
    a.fit.print_params(prms)
    for ant in fitants:
        for prm in shprms:
            prms[ant][prm] = prms[fitants[0]][prm]
    aa.set_params(prms)
    cat.set_params(prms)
    a1,a2,th = cat.get_srcshapes()
    #if n.all(asz == 0): asz = None  # Making None bypasses a computation step
    score,nsamples = 0.,0.
    cnt,curtime = 0,None
    for uvfile in args:
        sys.stdout.write('.') ; sys.stdout.flush()
        uv = a.miriad.UV(uvfile)
        a.scripting.uv_selector(uv, opts.ant, opts.pol)
        for (uvw,t,(i,j)),d,f in uv.all(raw=True):
            # Use only every Nth integration, if decimation is specified.
            if curtime != t:
                curtime = t
                cnt = (cnt + 1) % opts.decimate
                if cnt == 0:
                    aa.set_jultime(t)
                    cat.compute(aa)
                    eqs = cat.get_crds('eq', ncrd=3)
                    flx = cat.get_fluxes()
                    ind = cat.get_indices()
                    aa.sim_cache(eqs, flx, indices=ind,
                        mfreqs=mfq, srcshapes=(a1,a2,th))
            if cnt != 0: continue
            if not opts.sim_autos and i == j: continue
            d = d.take(chans)
            f = f.take(chans)
            sim_d = aa.sim(i, j, pol=a.miriad.pol2str[uv['pol']])
            sq,difsq = n.abs(d)**2, n.abs(d - sim_d)**2
            difsq = n.where(f, 0, difsq)
            if opts.norm:
                sq = n.ma.masked_equal(sq, 0)
                score += (difsq/sq).sum()
                nsamples += n.logical_not(f).sum()
            else:
                score += difsq.sum()
                nsamples += sq.sum()
    score = n.sqrt(score / nsamples)
    print
    if first_fit is None: first_fit = score
    print 'Score:', score, 
    print '(%2.2f%% of %f)' % (100 * score / first_fit, first_fit)
    print '-------------------------------------------------------------------'
    return score / first_fit

a.optimize.fmin(fit_func, prm_list,
    maxfun=n.Inf, maxiter=n.Inf, ftol=1e-100, xtol=1e-100)

'''
def anneal(func, x0, T_x, cooling=lambda i: 3*(n.cos(i/50.)+1), 
        maxiter=1000, verbose=True):
    score = func(x0)
    for i in range(maxiter):
        if verbose: print 'Step:', i, 'T fraction:', cooling(i)
        delta = n.random.normal(scale=1., size=x0.shape) * cooling(i) * T_x
        n_x = x0 + delta
        n_score = func(n_x)
        if n_score < score: x0, score = n_x, n_score
        if verbose:
            print 'Best score:', score
            print 'Best x:', x0
    return x0, score

prm_list = n.array(prm_list)
x0, score = anneal(fit_func, prm_list, .0001*prm_list)
print x0
print 'Score:', score
print 'Done!'
'''



#! /usr/bin/env python
"""
Models visibilities for various catalog sources and creates a new Miriad UV
file containing either the simulated data, or the residual when the model
is removed from measured data.

Author: Aaron Parsons
"""
import numpy as n, aipy as a, optparse, os, sys, ephem

o = optparse.OptionParser()
o.set_usage('mdlvis.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, loc=True, src=True)
o.add_option('--sim', dest='sim', action='store_true',
    help='Output a simulated dataset (rather than subtracting).')
o.add_option('-f', '--flag', dest='flag', action='store_true',
    help='If outputting a simulated data set, mimic the data flagging of the original dataset.')
o.add_option('-m', '--map', dest='map',
    help='The Healpix map to use for simulation input.')
o.add_option('--iepoch', dest='iepoch', default=ephem.J2000, 
    help='The epoch of coordinates in the map. Default J2000.')
o.add_option('--freq', dest='freq', default=.150, type='float',
    help='Frequency of flux data in map.')
o.add_option('-n', '--noiselev', dest='noiselev', default=0., type='float',
    help='RMS amplitude of noise added to each UV sample of simulation.')
o.add_option('--nchan', dest='nchan', default=256, type='float',
    help='Number of channels in simulated data if no input data to mimic.')
o.add_option('--sfreq', dest='sfreq', default=.75, type='float',
    help='Start frequency (GHz) in simulated data if no input data to mimic.')
o.add_option('--sdf', dest='sdf', default=.150/256, type='float',
    help='Channel spacing (GHz) in simulated data if no input data to mimic.')
o.add_option('--inttime', dest='inttime', default=10, type='float',
    help='Integration time (s) in simulated data if no input data to mimic.')
o.add_option('--startjd', dest='startjd', default=2454600., type='float',
    help='Julian Date to start observation if no input data to mimic.')
o.add_option('--endjd', dest='endjd', default=2454601., type='float',
    help='Julian Date to end observation if no input data to mimic.')
opts, args = o.parse_args(sys.argv[1:])

# Parse command-line options
uv = a.miriad.UV(args[0])
aa = a.loc.get_aa(opts.loc, uv['sdf'], uv['sfreq'], uv['nchan'])
p,d,f = uv.read(raw=True)
no_flags = n.zeros_like(f)
del(uv)

# Generate a model of the sky with point sources and a pixel map

mfq,a1s,a2s,ths = [], [], [], []
# Initialize point sources
if not opts.src is None:
    cat = a.scripting.parse_srcs(opts.src, force_cat=True)
    cat.set_params(a.loc.get_src_prms(opts.loc))
    mfq.append(cat.get_mfreqs())
    a1,a2,th = cat.get_srcshapes()
    a1s.append(a1); a2s.append(a2); ths.append(th)

# Initialize pixel map
if not opts.map is None:
    h = a.map.Map(fromfits=opts.map)
    px = n.arange(h.npix())
    try: mflx, i_poly = h[px]
    except(ValueError): mflx = h[px]
    px = n.compress(mflx > 0, px)
    try: mflx, i_poly = h[px]
    except(ValueError):
        mflx = h[px]
        i_poly = [n.zeros_like(mflx)]
    mind = i_poly[0]    # Only implementing first index term for now
    mmfq = opts.freq * n.ones_like(mind)
    mfq.append(mmfq)
    x,y,z = h.px2crd(px, ncrd=3)
    m_eq = n.array((x,y,z))
    # Should pixels include information for resolving them?
    a1s.append(n.zeros_like(mmfq))
    a2s.append(n.zeros_like(mmfq))
    ths.append(n.zeros_like(mmfq))
mfq = n.concatenate(mfq)
a1s = n.concatenate(a1s)
a2s = n.concatenate(a2s)
ths = n.concatenate(ths)
#if n.all(asz == 0): asz = None

# A pipe for just outputting the model
curtime = None
def mdl(uv, p, d, f):
    global curtime, eqs
    uvw, t, (i,j) = p
    if i == j: return p, d, f
    if curtime != t:
        curtime = t
        aa.set_jultime(t)
        eqs,flx,ind = [],[],[]
        if not opts.src is None:
            cat.compute(aa)
            eqs.append(cat.get_crds('eq', ncrd=3))
            flx.append(cat.get_fluxes())
            ind.append(cat.get_indices())
        if not opts.map is None:
            m_precess = a.coord.convert_m('eq','eq',
                iepoch=opts.iepoch, oepoch=aa.epoch)
            eqs.append(n.dot(m_precess, m_eq))
            flx.append(mflx); ind.append(mind)
        eqs = n.concatenate(eqs, axis=-1)
        flx = n.concatenate(flx)
        ind = n.concatenate(ind)
        aa.sim_cache(eqs, flx, indices=ind, mfreqs=mfq, srcshapes=(a1,a2,th))
    sd = aa.sim(i, j, pol=a.miriad.pol2str[uv['pol']])
    if opts.sim:
        d = sd
        if not opts.flag: f = no_flags
    else: d -= sd
    if opts.noiselev != 0:
        # Add on some noise for a more realistic experience
        noise_amp = n.random.random(d.shape) * opts.noiselev
        noise_phs = n.random.random(d.shape) * 2*n.pi * 1j
        noise = noise_amp * n.exp(noise_phs)
        d += noise
    return p, d, f

# Run mdl on all files
for filename in args:
    print filename
    uvofile = filename + 's'
    if os.path.exists(uvofile):
        print 'File exists: skipping'
        continue
    uvi = a.miriad.UV(filename)
    uvo = a.miriad.UV(uvofile, status='new')
    uvo.init_from_uv(uvi)
    uvo.pipe(uvi, mfunc=mdl, raw=True)




#! /usr/bin/env python
"""
Rotate zenith UV data to a particular source.  

Author: Aaron Parsons
"""

import aipy as a, numpy as n, sys, os, optparse

o = optparse.OptionParser()
o.set_usage('phs2src.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, loc=True, src=True)
o.add_option('--setphs', dest='setphs', action='store_true',
    help='Instead of rotating phase, assign a phase corresponding to the specified source.')
opts,args = o.parse_args(sys.argv[1:])

# Parse command-line options
uv = a.miriad.UV(args[0])
aa = a.loc.get_aa(opts.loc, uv['sdf'], uv['sfreq'], uv['nchan'])
src = a.scripting.parse_srcs(opts.src)
del(uv)

# A pipe to use for phasing to a source
curtime,cache,top = None, {}, None
def phs(uv, p, d, f):
    global curtime, cache, top
    uvw, t, (i,j) = p
    if curtime != t:
        curtime = t
        aa.set_jultime(t)
        src.compute(aa)
    if i == j: return p, d, f
    try:
        if opts.setphs: d = aa.unphs2src(n.abs(d), src, i, j)
        else: d = aa.phs2src(d, src, i, j)
    except(a.ant.PointingError): d *= 0
    return p, d, f

# Process data
for filename in args:
    print filename
    uvofile = filename + '.' + opts.src
    if os.path.exists(uvofile):
        print 'File exists: skipping'
        continue
    uvi = a.miriad.UV(filename)
    uvo = a.miriad.UV(uvofile, status='new')
    uvo.init_from_uv(uvi)
    uvo.pipe(uvi, mfunc=phs, raw=True)




#! /usr/bin/env python
"""
A script for removing a source using a fringe-rate/delay transform.

Author: Aaron Parsons
"""

import aipy as a, numpy as n, os, sys, optparse

o = optparse.OptionParser()
o.set_usage('filter_src.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, src=True, loc=True)
o.add_option('-f', '--fng_w', dest='fng_w', type=int, default=-1,
    help='The number of fringe bins to null.  Default is -1 = no fringe filtering.')
o.add_option('-d', '--dly_w', dest='dly_w', type=int, default=5,
    help='The number of delay bins to null.')
o.add_option('-e', '--extract', dest='extract', action='store_true',
    help='Extract the source instead of removing it.')
o.add_option('-t', '--together', dest='together', action='store_true',
    help='Perform fringe-rate transform over all files, not file-by-file.')
o.add_option('--clean', dest='clean', type='float', default=1e-3,
    help='Deconvolve delay-domain data by the "beam response" that results from flagged data.  Specify a tolerance for termination (usually 1e-2 or 1e-3).')
opts, args = o.parse_args(sys.argv[1:])

uv = a.miriad.UV(args[0])
if not opts.src is None:
    aa = a.loc.get_aa(opts.loc, uv['sdf'], uv['sfreq'], uv['nchan'])
    src = a.scripting.parse_srcs(opts.src)
else:
    opts.src = 'cut'
    src = None
del(uv)

phs_dat = {}
for uvfile in args:
    print 'Working on', uvfile
    if not opts.together: phs_dat = {}
    if opts.extract: uvofile = uvfile+'.e_'+opts.src
    else: uvofile = uvfile+'.'+opts.src
    if os.path.exists(uvofile):
        if opts.together:
            print uvofile, 'exists, aborting.'
            sys.exit(0)
        print uvofile, 'exists, skipping.'
        continue
    uvi = a.miriad.UV(uvfile)

    curtime = None
    print '    Performing delay transform and cleaning...'
    for (uvw,t,(i,j)),d,f in uvi.all(raw=True):
        if i == j: continue
        if curtime != t:
            curtime = t
            if not src is None:
                aa.set_jultime(t)
                src.compute(aa)
        bl = a.miriad.ij2bl(i,j)
        try:
            flags = n.logical_not(f).astype(n.float)
            gain = n.sqrt(n.average(flags**2))
            ker = n.fft.ifft(flags)
            d = n.where(f, 0, d)
            if not src is None: d = aa.phs2src(d, src, i, j)
            d = n.fft.ifft(d)
            if not n.all(d == 0):
                d, info = a.deconv.clean1d(d, ker, tol=opts.clean)
                d += info['res'] / gain
        except(a.ant.PointingError): d = n.zeros_like(d)
        try: phs_dat[bl].append(d)
        except(KeyError): phs_dat[bl] = [d]

    if not opts.together:
        print '    Performing fringe rate transform and cleaning...'
        for bl in phs_dat:
            d = n.array(phs_dat[bl])
            if opts.fng_w != -1:
                flags = n.where(d[:,0] != 0, 1., 0.)
                gain = n.sqrt(n.average(flags**2))
                ker = n.fft.ifft(flags)
                d = n.fft.ifft(d, axis=0)
                for chan in range(d.shape[1]):
                    d[:,chan],info = a.deconv.clean1d(d[:,chan],ker,tol=opts.clean)
                    d[:,chan] += info['res'] / gain
                x1, x2 = opts.fng_w, -opts.fng_w+1
                if x2 == 0: x2 = d.shape[0]
                d[x1:x2,:] = 0
            y1, y2 = opts.dly_w, -opts.dly_w
            if y2 == 0: y2 = d.shape[1]
            d[:,y1:y2] = 0
            if opts.fng_w != -1: d = n.fft.fft(d, axis=0)
            d = n.fft.fft(d, axis=1)
            phs_dat[bl] = d

        cnt = {}
        for bl in phs_dat: cnt[bl] = 0

        # Generate a pipe for removing average phase bias from data
        def rm_mfunc(uv, p, d):
            uvw, t, (i,j) = p
            if i == j: return p, d
            bl = a.miriad.ij2bl(i,j)
            if not src is None:
                aa.set_jultime(t)
                src.compute(aa)
            data = phs_dat[bl][cnt[bl],:]
            if not src is None:
                try: data = aa.unphs2src(data, src, i, j)
                except(a.ant.PointingError):
                    if opts.extract: d *= 0
                    data = 0
            cnt[bl] += 1
            if opts.extract: return p, n.ma.array(data, mask=d.mask)
            else: return p, d - data

        print '    Writing out filtered data...'
        # Apply the pipe to the data
        uvi.rewind()
        uvo = a.miriad.UV(uvofile, status='new')
        uvo.init_from_uv(uvi)
        uvo.pipe(uvi, mfunc=rm_mfunc, append2hist='FILTER_SRC: src=%s fng_w=%d dly_w=%d extract=%s clean=%f\n' % (opts.src, opts.fng_w, opts.dly_w, opts.extract, opts.clean))

if opts.together:
    print '    Performing fringe rate transform and cleaning...'
    for bl in phs_dat:
        d = n.array(phs_dat[bl])
        if opts.fng_w != -1:
            flags = n.where(d[:,0] != 0, 1., 0.)
            gain = n.sqrt(n.average(flags**2))
            ker = n.fft.ifft(flags)
            d = n.fft.ifft(d, axis=0)
            for chan in range(d.shape[1]):
                d[:,chan],info = a.deconv.clean1d(d[:,chan],ker,tol=opts.clean)
                d[:,chan] += info['res'] / gain
            x1, x2 = opts.fng_w, -opts.fng_w+1
            if x2 == 0: x2 = d.shape[0]
            d[x1:x2,:] = 0
        y1, y2 = opts.dly_w, -opts.dly_w
        if y2 == 0: y2 = d.shape[1]
        d[:,y1:y2] = 0
        if opts.fng_w != -1: d = n.fft.fft(d, axis=0)
        d = n.fft.fft(d, axis=1)
        phs_dat[bl] = d

    cnt = {}
    for bl in phs_dat: cnt[bl] = 0

    # Generate a pipe for removing average phase bias from data
    def rm_mfunc(uv, p, d):
        uvw, t, (i,j) = p
        if i == j: return p, d
        bl = a.miriad.ij2bl(i,j)
        if not src is None:
            aa.set_jultime(t)
            src.compute(aa)
        data = phs_dat[bl][cnt[bl],:]
        if not src is None:
            try: data = aa.unphs2src(data, src, i, j)
            except(a.ant.PointingError):
                if opts.extract: d *= 0
                data = 0
        cnt[bl] += 1
        if opts.extract: return p, n.ma.array(data, mask=d.mask)
        else: return p, d - data

    for uvfile in args:
        print 'Working on', uvfile
        if not opts.together: phs_dat = {}
        if opts.extract: uvofile = uvfile+'.e_'+opts.src
        else: uvofile = uvfile+'.'+opts.src
        uvi = a.miriad.UV(uvfile)
        uvo = a.miriad.UV(uvofile, status='new')
        uvo.init_from_uv(uvi)
        uvo.pipe(uvi, mfunc=rm_mfunc, append2hist='FILTER_SRC: src=%s fng_w=%d dly_w=%d extract=%s clean=%f\n' % (opts.src, opts.fng_w, opts.dly_w, opts.extract, opts.clean))



"""
Module adding data simulation support to AntennaArrays.  For most classes,
this means adding gain/amplitude information (as a function of frequency).
"""

import ant, numpy as n, ephem, coord, healpix

#  ____           _ _       ____            _       
# |  _ \ __ _  __| (_) ___ | __ )  ___   __| |_   _ 
# | |_) / _` |/ _` | |/ _ \|  _ \ / _ \ / _` | | | |
# |  _ < (_| | (_| | | (_) | |_) | (_) | (_| | |_| |
# |_| \_\__,_|\__,_|_|\___/|____/ \___/ \__,_|\__, |
#                                             |___/ 

class RadioBody:
    """Class defining flux and spectral index of a celestial source."""
    def __init__(self, janskies, mfreq, index, srcshape):
        """janskies = source strength. Can be polynomial in hour angle (-pi,pi)
        mfreq = frequency (in GHz) where strength was measured
        index = index of power-law spectral model of source emission.  Can be
        polynomial in hour angle."""
        try: len(janskies)
        except: janskies = [janskies]
        try: len(index)
        except: index = [index]
        self._janskies = janskies
        self.mfreq = mfreq
        self._index = index
        try: len(srcshape)
        except(TypeError): srcshape = [srcshape, srcshape, 0.]
        assert(len(srcshape) == 3)
        self.srcshape = list(srcshape)
    def compute(self, observer):
        """Update fluxes relative to the provided observer.  Must be
        called at each time step before accessing information."""
        self.janskies = n.clip(n.polyval(self._janskies, 
            (observer.sidereal_time()-self.ra+n.pi) % (2*n.pi) - n.pi), 
            0, n.Inf)
        self.index = n.polyval(self._index, observer.sidereal_time())

#  ____           _ _       _____ _              _ ____            _       
# |  _ \ __ _  __| (_) ___ |  ___(_)_  _____  __| | __ )  ___   __| |_   _ 
# | |_) / _` |/ _` | |/ _ \| |_  | \ \/ / _ \/ _` |  _ \ / _ \ / _` | | | |
# |  _ < (_| | (_| | | (_) |  _| | |>  <  __/ (_| | |_) | (_) | (_| | |_| |
# |_| \_\__,_|\__,_|_|\___/|_|   |_/_/\_\___|\__,_|____/ \___/ \__,_|\__, |
#                                                                    |___/ 

class RadioFixedBody(ant.RadioFixedBody, RadioBody):
    """Class representing a source at fixed RA,DEC.  Adds flux information to
    ant.RadioFixedBody."""
    def __init__(self, ra, dec, janskies, mfreq=.150, 
            index=-1., srcshape=(0.,0.,0.), name='', **kwargs):
        """ra = source's right ascension (epoch=J2000)
        dec = source's declination (epoch=J2000)
        janskies = source strength. Can be polynomial in hour angle (-pi,pi)
        mfreq = frequency (in GHz) where strength was measured
        index = index of power-law spectral model of source emission.  Can be
        polynomial in hour angle."""
        ant.RadioFixedBody.__init__(self, ra, dec, name=name)
        RadioBody.__init__(self, janskies, mfreq, index, srcshape)
    def compute(self, observer):
        ant.RadioFixedBody.compute(self, observer)
        RadioBody.compute(self, observer)

#  ____           _ _      ____                  _       _ 
# |  _ \ __ _  __| (_) ___/ ___| _ __   ___  ___(_) __ _| |
# | |_) / _` |/ _` | |/ _ \___ \| '_ \ / _ \/ __| |/ _` | |
# |  _ < (_| | (_| | | (_) |__) | |_) |  __/ (__| | (_| | |
# |_| \_\__,_|\__,_|_|\___/____/| .__/ \___|\___|_|\__,_|_|
#                               |_|                        

class RadioSpecial(ant.RadioSpecial, RadioBody):
    """Class representing moving sources (Sun,Moon,planets).  Adds flux
    information to ant.RadioSpecial."""
    def __init__(self, name, janskies, mfreq=.150, index=-1.,
            srcshape=(0.,0.,0.), **kwargs):
        """janskies = source strength. Can be polynomial in hour angle (-pi,pi)
        mfreq = frequency (in GHz) where strength was measured
        index = index of power-law spectral model of source emission.  Can be
        polynomial in hour angle."""
        ant.RadioSpecial.__init__(self, name)
        RadioBody.__init__(self, janskies, mfreq, index, srcshape)
    def compute(self, observer):
        ant.RadioSpecial.compute(self, observer)
        RadioBody.compute(self, observer)

#  ____            ____      _        _             
# / ___| _ __ ___ / ___|__ _| |_ __ _| | ___   __ _ 
# \___ \| '__/ __| |   / _` | __/ _` | |/ _ \ / _` |
#  ___) | | | (__| |__| (_| | || (_| | | (_) | (_| |
# |____/|_|  \___|\____\__,_|\__\__,_|_|\___/ \__, |
#                                             |___/ 

class SrcCatalog(ant.SrcCatalog):
    """Class for holding a catalog of celestial sources."""
    def get_fluxes(self, srcs=None):
        """Return list of fluxes of all src objects in catalog."""
        if srcs is None: srcs = self.keys()
        return n.array([self[s].janskies for s in srcs])
    def get_indices(self, srcs=None):
        """Return list of spectral indices of all src objects in catalog."""
        if srcs is None: srcs = self.keys()
        return n.array([self[s].index for s in srcs])
    def get_mfreqs(self, srcs=None):
        """Return list of frequencies where strength is measured for all src 
        objects in catalog."""
        if srcs is None: srcs = self.keys()
        return n.array([self[s].mfreq for s in srcs])
    def get_srcshapes(self, srcs=None):
        """Return list of angular sizes of all src objects in catalog."""
        if srcs is None: srcs = self.keys()
        a1 = n.array([self[s].srcshape[0] for s in srcs])
        a2 = n.array([self[s].srcshape[1] for s in srcs])
        th = n.array([self[s].srcshape[2] for s in srcs])
        return (a1,a2,th)

#  ____
# | __ )  ___  __ _ _ __ ___
# |  _ \ / _ \/ _` | '_ ` _ \
# | |_) |  __/ (_| | | | | | |
# |____/ \___|\__,_|_| |_| |_|

class BeamFlat(ant.Beam):
    """Representation of a flat (gain=1) antenna beam pattern."""
    def response(self, xyz):
        x,y,z = n.array(xyz)
        return n.ones((self.afreqs.size, x.size))

class Beam2DGaussian(ant.Beam):
    """Representation of a 2D Gaussian beam pattern, with default setting for 
    a flat beam."""
    def __init__(self, freqs, xwidth=n.Inf, ywidth=n.Inf):
        """xwidth = angular width (radians) in EW direction
        ywidth = angular width (radians) in NS direction"""
        ant.Beam.__init__(self, freqs)
        self.update(xwidth=xwidth, ywidth=ywidth)
    def update(self, xwidth=None, ywidth=None):
        """Set the width in the x (EW) and y (NS) directions of the gaussian 
        beam."""
        if not xwidth is None: self.xwidth = xwidth
        if not ywidth is None: self.ywidth = ywidth
    def response(self, xyz):
        """Return beam response across active band for specified topocentric 
        coordinates: (x=E,y=N,z=UP). x,y,z may be arrays of multiple 
        coordinates.  Returns 'x' linear polarization (rotate pi/2 for 'y')."""
        x,y,z = xyz
        x,y = n.arcsin(x)/self.xwidth, n.arcsin(y)/self.ywidth
        resp = n.sqrt(n.exp(-(x**2 + y**2)))
        resp = n.resize(resp, (self.afreqs.size, resp.size))
        return resp

class BeamPolynomial(ant.Beam):
    """Representation of a gaussian beam model whose width varies with azimuth
    angle and with frequency."""
    def __init__(self, freqs, poly_azfreq=n.array([[.5]])):
        """poly_azfreq = a 2D polynomial in cos(2*n*az) for first axis and 
        in freq**n for second axis."""
        self.poly = poly_azfreq
        ant.Beam.__init__(self, freqs)
        self.update(poly_azfreq)
    def select_chans(self, active_chans):
        """Select only enumerated channels to use for future calculations."""
        ant.Beam.select_chans(self, active_chans)
        self.update()
    def update(self, poly_azfreq=None):
        """Update beam with new polynomial coefficients.
        poly_azfreq = a 2D polynomial in cos(2*n*az) for first axis and 
        in freq**n for second axis."""
        if poly_azfreq is None: poly_azfreq = self.poly
        elif len(poly_azfreq.shape) == 1: poly_azfreq.shape = self.poly.shape
        self.poly = poly_azfreq
        f = n.resize(self.afreqs, (self.poly.shape[1], self.afreqs.size))
        f = f**n.array([range(self.poly.shape[1])]).transpose()
        self.sigma = n.dot(self.poly, f)
    def response(self, top):
        """Return beam response across active band for specified topocentric 
        coordinates (x=E,y=N,z=UP). x,y,z may be multiple coordinates.  
        Returns 'x' pol (rotate pi/2 for 'y')."""
        az,alt = coord.top2azalt(top)
        zang = n.pi/2 - alt
        if zang.size == 1:
            zang = n.array([zang]); zang.shape = (1,)
            az = n.array([az]); az.shape = (1,)
        a = 2 * n.arange(self.poly.shape[0], dtype=n.float)
        a.shape = (1,) + a.shape; az.shape += (1,); zang.shape += (1,)
        a = n.cos(n.dot(az, a))
        a[:,0] = 0.5
        s = n.dot(a, self.sigma)
        return n.sqrt(n.exp(-(zang/s)**2)).transpose()

class BeamAlm(ant.Beam):
    """Representation of a beam model where each pointing has a response
    defined as a polynomial in frequency, and the spatial distributions of 
    these coefficients decomposed into spherical harmonics."""
    def __init__(self, freqs, lmax=8, mmax=8, deg=7, nside=64, coeffs={}):
        """lmax = maximum spherical harmonic term
        mmax = maximum spherical harmonic term in the z direction
        deg = order of polynomial to used for mapping response of each pointing
        nside = resolution of underlying HealpixMap to use
        coeffs = dictionary of polynomial term (integer) and corresponding Alm 
        coefficients (see healpix.py doc)."""
        self.alm = [healpix.Alm(lmax,mmax) for i in range(deg+1)]
        self.hmap = [healpix.HealpixMap(nside,scheme='RING',interp=True)
            for a in self.alm]
        ant.Beam.__init__(self, freqs)
        self.update(coeffs)
    def update(self, coeffs={}):
        """Update beam model using new set of coefficients.
        coeffs = dictionary of polynomial term (integer) and corresponding Alm 
        coefficients (see healpix.py doc)."""
        for c in coeffs:
            if c >= len(self.alm): continue
            self.alm[-1-c].set_data(coeffs[c])
            self.hmap[-1-c].from_alm(self.alm[-1-c])
    def response(self, top):
        """Return beam response across active band for specified topocentric 
        coordinates (x=E,y=N,z=UP). x,y,z may be multiple coordinates.  
        Returns 'x' pol (rotate pi/2 for 'y')."""
        top = [healpix.mk_arr(c, dtype=n.double) for c in top]
        px,wgts = self.hmap[0].crd2px(*top, **{'interpolate':1})
        poly = n.array([n.sum(h.map[px] * wgts, axis=-1) for h in self.hmap])
        rv = n.polyval(poly, n.reshape(self.afreqs, (self.afreqs.size, 1)))
        return rv

#     _          _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` |
#  / ___ \| | | | ||  __/ | | | | | | (_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_|

class Antenna(ant.Antenna):
    """Representation of physical location and beam pattern of individual 
    antenna in array."""
    def __init__(self, x, y, z, beam, delay=0., offset=0., bp_r=n.array([1]),
            bp_i=n.array([0]), amp=1, pointing=(0.,n.pi/2,0), **kwargs):
        """x,y z = antenna coordinates in equatorial (ns) coordinates
        beam = Beam object (implements response() function)
        delay = signal delay (linear phase vs. frequency)
        offset = signal phase offset (constant phase vs. frequency)
        bp_r = polynomial (in freq) modeling real component of passband
        bp_i = polynomial (in freq) modeling imaginary component of passband
        amp = overall multiplicative scaling of gain
        pointing = antenna pointing (az,alt).  Default is zenith."""
        ant.Antenna.__init__(self, x,y,z, beam=beam, delay=delay, offset=offset)
        self.update_gain(bp_r, bp_i, amp)
        self.update_pointing(*pointing)
    def select_chans(self, active_chans=None):
        """Select only enumerated channels to use for future calculations."""
        ant.Antenna.select_chans(self, active_chans)
        self.update_gain()
    def update_gain(self, bp_r=None, bp_i=None, amp=None):
        """Update passband information using provided parameters.
        bp_r = polynomial (in freq) modeling real component of passband
        bp_i = polynomial (in freq) modeling imaginary component of passband
        amp = overall multiplicative scaling of gain"""
        if not bp_r is None:
            try: len(bp_r)
            except: bp_r = [bp_r]
            self.bp_r = bp_r
        if not bp_i is None:
            try: len(bp_i)
            except: bp_i = [bp_i]
            self.bp_i = bp_i
        if not amp is None: self.amp = amp
        bp = n.polyval(self.bp_r, self.beam.afreqs) + \
             1j*n.polyval(self.bp_i, self.beam.afreqs)
        self._gain = self.amp * bp
    def update_pointing(self, az=0, alt=n.pi/2, twist=0):
        """Set the antenna beam to point at (az, alt) with specified
        right-hand twist to polarizations.  Polarization y is assumed
        to be +pi/2 azimuth from pol x."""
        y, z = n.array([0,1,0]), n.array([0,0,1])
        rot = coord.rot_m(twist, z)
        rot = n.dot(rot, coord.rot_m(alt-n.pi/2, y))
        rot = n.dot(rot, coord.rot_m(-az, z))
        self.rot_pol_x = rot
        self.rot_pol_y = n.dot(coord.rot_m(-n.pi/2, z), rot)
    def bm_response(self, top, pol='x'):
        """Return response of beam for specified polarization."""
        top = n.array(top)
        top = {'x':top, 'y':n.dot(self.rot_pol_y, top)}[pol]
        x,y,z = top
        return self.beam.response((x,y,z))
    def passband(self, conj=False):
        if conj: return n.conjugate(self._gain)
        else: return self._gain

#     _          _                            _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _   / \   _ __ _ __ __ _ _   _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` | / _ \ | '__| '__/ _` | | | |
#  / ___ \| | | | ||  __/ | | | | | | (_| |/ ___ \| |  | | | (_| | |_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_/_/   \_\_|  |_|  \__,_|\__, |
#                                                                 |___/ 

class AntennaArray(ant.AntennaArray):
    """Representation of location and time of observation, and response of
    array of antennas as function of pointing and frequency."""
    def set_jultime(self, t=None):
        """Set current time as a Julian date."""
        ant.AntennaArray.set_jultime(self, t=t)
        self.eq2top_m = coord.eq2top_m(-self.sidereal_time(), self.lat)
        self._cache = None
    def passband(self, i, j):
        return self.ants[i].passband() * self.ants[j].passband(conj=True)
    def bm_response(self, i, j, pol='xx'):
        assert(pol in ('xx','yy','xy','yx'))
        p1, p2 = pol
        # Check that we have cached results needed.  If not, cache them.
        for c,p in zip([i,j], [p1,p2]):
            if not self._cache.has_key(c): self._cache[c] = {}
            if not self._cache[c].has_key(p):
                x,y,z = self._cache['s_top']
                resp = self.ants[c].bm_response((x,y,z), pol=p).transpose()
                self._cache[c][p] = resp
        return self._cache[i][p1] * n.conjugate(self._cache[j][p2])
    def sim_cache(self, s_eqs, fluxes=n.array([1.]), indices=n.array([0.]), 
            mfreqs=n.array([.150]), srcshapes=(0,0,0)):
        """Cache intermediate computations given catalog information to speed
        simulation for multiple baselines.  For efficiency, should only be 
        called once per time setting.  MUST be called before sim().
        s_eqs = array of equatorial vectors for all celestial sources
        fluxes = array of fluxes for all celestial sources
        indices = array of spectral indices of all celestial sources
        mfreqs = array of frequencies where fluxes were measured
        srcshapes = (a1,a2,th) where a1,a2 are angular sizes along the 
            semimajor, semiminor axes, and th is the angle (in radians) of
            the semimajor axis from E."""
        # Get topocentric coordinates of all srcs
        src_top = n.dot(self.eq2top_m, s_eqs)
        # Throw out everything that is below the horizon
        valid = n.logical_and(src_top[2,:] > 0, fluxes > 0)
        if n.all(valid == 0): self._cache = {}
        else:
            fluxes = fluxes.compress(valid)
            indices = indices.compress(valid)
            mfreqs = mfreqs.compress(valid)
            a1,a2,th = srcshapes
            try:
                a1 = a1.compress(valid)
                a2 = a2.compress(valid)
                th = th.compress(valid)
            except(AttributeError): pass
            src_top = src_top.compress(valid, axis=1)
            s_eqs = s_eqs.compress(valid, axis=1)
            # Get src fluxes vs. freq
            fluxes.shape = (fluxes.size, 1)
            mfreqs.shape = (mfreqs.size, 1)
            indices.shape = (indices.size, 1)
            freqs = n.resize(self.ants[0].beam.afreqs, 
                (fluxes.size, self.ants[0].beam.afreqs.size))
            self._cache = {
                'I_sf':fluxes * (freqs / mfreqs)**indices,
                's_eqs': s_eqs,
                's_top': src_top,
                's_shp': (a1,a2,th),
            }
    def sim(self, i, j, pol='xx'):
        """Simulate visibilites for the specified (i,j) baseline and 
        polarization.  sim_cache() must be called at each time step before 
        this will return valid results."""
        assert(pol in ('xx','yy','xy','yx'))
        if self._cache is None:
            raise RuntimeError('sim_cache() must be called before the first sim() call at each time step.')
        elif self._cache == {}:
            return n.zeros_like(self.passband(i,j))
        s_eqs = self._cache['s_eqs']
        I_sf = self._cache['I_sf'] * \
            self.resolve_src(s_eqs, i, j, self._cache['s_shp'])
        Gij_sf = self.passband(i,j)
        Bij_sf = self.bm_response(i,j,pol=pol)
        if len(Bij_sf.shape) == 2: Gij_sf = n.reshape(Gij_sf, (1, Gij_sf.size))
        # Get the phase of each src vs. freq, also does resolution effects
        E_sf = n.conjugate(self.gen_phs(s_eqs, i, j))
        try: E_sf.shape = I_sf.shape
        except(AttributeError): pass
        # Combine and sum over sources
        GBIE_sf = Gij_sf * Bij_sf * I_sf * E_sf
        Vij_f = GBIE_sf.sum(axis=0)
        return Vij_f



"""
Module containing utilities (like parsing of certain command-line arguments) 
for writing scripts.
"""

import miriad, ant, sim, fit, src, numpy as n

def add_standard_options(optparser, ant=False, pol=False, chan=False, 
        loc=False, loc2=False, src=False, dec=False):
    """Add standard command-line options to an optparse.OptionParser() on an 
    opt in basis (i.e. specify =True for each option to be added)."""
    if ant: optparser.add_option ('-a', '--ant', dest='ant', default='cross',
         help='Select antennas/baselines to include.  Options are "all", "auto", "cross", "<ant1 #>_<ant2 #>,..." (a list of baselines), or "<ant1 #>,..." (a list of antennas).  Default is "cross".')
    if pol: optparser.add_option('-p', '--pol', dest='pol', 
        help='Choose polarization (xx, yy, xy, yx) to include.')
    if chan: optparser.add_option('-c', '--chan', dest='chan', default='all',
        help='Select channels (taken after any delay/fringe transforms) to include.  Options are "all", "<chan1 #>,..." (a list of channels), or "<chan1 #>_<chan2 #>" (a range of channels).  Default is "all".')
    if loc: optparser.add_option('-l', '--loc', dest='loc', 
        help='Use specified <loc>.py for location-specific calibration.')
    elif loc2: optparser.add_option('-l', '--loc', dest='loc', action='append',
        help='Use specified <loc>.py for location-specific calibration.  Can intersperse locations with UV files to use different calibrations for different files.')
    if src: optparser.add_option('-s', '--src', dest='src',
        help='Phase centers/source catalog entries to use.  Options are "all", "<src_name1>,...", or "<ra XX[:XX:xx]>_<dec XX[:XX:xx]>".')
    if dec: optparser.add_option('-x', '--decimate', dest='decimate', 
        default=1, type='int',
        help='Use only every Nth integration.  Default is 1.')

def _strarg_to_range(strarg):
    """Split command-line lists/ranges into a list of numbers."""
    strarg = strarg.split(',')
    return [map(float, option.split('_')) for option in strarg]

def uv_selector(uv, ant_str, pol_str):
    """Call uv.select with appropriate options based on string argument for
    antennas (can be 'all', 'auto', 'cross', '0,1,2', or '0_1,0_2') and
    string for polarization ('xx','yy','xy','yx')."""
    if ant_str.startswith('all'): pass
    elif ant_str.startswith('auto'): uv.select('auto', 0, 0)
    elif ant_str.startswith('cross'): uv.select('auto', 0, 0, include=0)
    else:
        antopt = _strarg_to_range(ant_str)
        antopt1 = [ao[0] for ao in antopt if len(ao) == 1]
        antopt2 = [ao for ao in antopt if len(ao) == 2]
        for opt in antopt2:
            a1,a2 = opt
            uv.select('antennae', a1, a2)
        if len(antopt1) == 1: uv.select('antennae', antopt1[0], -1)
        else:
            for a1 in antopt1:
                for a2 in antopt1:
                    if a1 != a2: uv.select('antennae', a1, a2)
    try: polopt = miriad.str2pol[pol_str]
    except(KeyError): raise ValueError('--pol argument invalid or absent')
    uv.select('polarization', polopt, 0)

def parse_chans(chan_str, nchan):
    """Return array of active channels based on number of channels and
    string argument for chans (can be 'all', '20_30', or '55,56,57')."""
    if chan_str.startswith('all'): chans = n.arange(nchan)
    else:
        chanopt = _strarg_to_range(chan_str)
        if len(chanopt[0]) != 1:
            chanopt = [n.arange(x,y, dtype=n.int) for x,y in chanopt]
        chans = n.concatenate(chanopt)
    return chans.astype(n.int)

def parse_srcs(src_str, force_cat=False):
    """Return src/catalog based on string argument for src (can be "all",
    "<src_name1>,...", or "<ra XX[:XX:xx]>_<dec XX[:XX:xx]>").  Can force
    a single source to be returned as catalog with force_cat."""
    if src_str.startswith('all'): return src.get_catalog()
    src_opt = src_str.split(',')
    if len(src_opt) == 1:
        src_opt = src_opt[0].split('_')
        if len(src_opt) == 1: s = src.get_src(src_opt[0])
        else:
            ra,dec = src_opt
            s = fit.RadioFixedBody(ra, dec, 0, name=src_str)
        if force_cat: return src.get_catalog([s])
        else: return s
    else:
        return src.get_catalog(src_opt)

def files_to_locs(locations, uvfiles, sysargv):
    import optparse
    o = optparse.OptionParser(); add_standard_options(o, loc2=True)
    locs = {}
    pos = [sysargv.index(uv) for uv in uvfiles]
    curloc = locations[0]
    for i in range(len(pos)):
        locs[curloc] = locs.get(curloc, []) + [sysargv[pos[i]]]
        if i < len(pos)-1 and pos[i+1] - pos[i] > 1:
            opts,args = o.parse_args(sysargv[pos[i]+1:pos[i+1]])
            if not opts.loc is None:
                if curloc is None:
                    locs[opts.loc[-1]] = locs[None]
                    del(locs[None])
                curloc = opts.loc[-1]
    return locs




#! /usr/bin/env python
"""
This is a general-purpose script for making images from MIRIAD UV files.  Data
(optionally selected for baseline, channel) are read from the file, phased
to a provided position, normalized for passband/primary beam effects, gridded
to a UV matrix, imaged, and optionally deconvolved by a corresponding PSF to
produce a clean image.

Author: Aaron Parsons
"""

import aipy as a, numpy as n, sys, optparse, ephem, os

o = optparse.OptionParser()
o.set_usage('mk_img.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, ant=True, pol=True, chan=True, loc2=True,
    src=True, dec=True)
o.add_option('-r', '--residual', dest='residual', action='store_true',
    help='Display only the residual in the 4th panel (otherwise display sum of clean image and residual).')
o.add_option('-d', '--deconv', dest='deconv', default='mem',
    help='Attempt to deconvolve the dirty image by the dirty beam using the specified deconvolver (none,mem,lsq,cln,ann).')
o.add_option('-o', '--output', dest='output', default='bim',
    help='Comma delimited list of data to generate FITS files for.  Can be: dim (dirty image), dbm (dirty beam), cim (clean image), rim (residual image), bim (best image = clean + residuals), uvs (uv sampling), or bms (beam sampling).  Default is bim.')
o.add_option('--npts', dest='npts', type='int', default=200,
    help='If no src is provided, facet the sphere into this many pointings for making a map.  Default 200.')
o.add_option('--prefix', dest='prefix', default='im_',
    help='A string to prefix to all filenames.')
o.add_option('--fmt', dest='fmt', default='%04d',
    help='A format string for counting successive images written to files.  Default is %04d (i.e. 0001).')
o.add_option('--var', dest='var', type='float', default=.6,
    help='Starting guess for variance in maximum entropy fit (defaults to variance of dirty image.')
o.add_option('--tol', dest='tol', type='float', default=1e-6,
    help='Tolerance for successful deconvolution.  For annealing, interpreted as cooling speed.')
o.add_option('--maxiter', dest='maxiter', type='int', default=200,
    help='Number of allowable iterations per deconvolve attempt.')
o.add_option('-u', '--uniform', dest='uniform', type='float', default=0,
    help="Use uniform (rather than natural) weighting for uv bins that have a weight above the specified fraction of the maximum weighting.")
o.add_option('--skip_amp', dest='skip_amp', action='store_true',
    help='Do not use amplitude information to normalize visibilities.')
o.add_option('--skip_bm', dest='skip_bm', action='store_true',
    help='Do not weight visibilities by the strength of the primary beam.')
o.add_option('--size', dest='size', type='int', default=200,
    help='Size of maximum UV baseline.')
o.add_option('--res', dest='res', type='float', default=0.5,
    help='Resolution of UV matrix.')
o.add_option('--no_w', dest='no_w', action='store_true',
    help="Don't use W projection.")
o.add_option('--buf_thresh', dest='buf_thresh', default=1.8e6, type='float',
    help='Maximum amount of data to buffer before gridding.  Excessive gridding takes performance hit, but if buffer exceeds memory available... ouch.')
opts, args = o.parse_args(sys.argv[1:])

# Parse command-line options
locs = a.scripting.files_to_locs(opts.loc, args, sys.argv)
aas = {}
for L in locs:
    uv = a.miriad.UV(locs[L][0])
    (j,t,j),j = uv.read()
    chans = a.scripting.parse_chans(opts.chan, uv['nchan'])
    a.scripting.uv_selector(uv, opts.ant, opts.pol)
    aa = a.loc.get_aa(L, uv['sdf'], uv['sfreq'], uv['nchan'])
    aa.select_chans(chans)
    aas[L] = aa
    afreqs = aa.ants[0].beam.afreqs
    cfreq = n.average(afreqs)
    aa.set_jultime(t)
outputs = opts.output.split(',')

if not opts.src is None: cat = a.scripting.parse_srcs(opts.src, force_cat=True)
else:
    ras,decs = a.map.facet_centers(opts.npts, ncrd=2)
    srcs = [a.ant.RadioFixedBody(ra,dec,name=str(i)) 
        for i,(ra,dec) in enumerate(zip(ras,decs))]
    cat = a.ant.SrcCatalog(srcs)

if opts.no_w: im = a.img.Img(opts.size, opts.res, mf_order=0)
else: im = a.img.ImgW(opts.size, opts.res, mf_order=0)
DIM = int(opts.size/opts.res)
del(uv)

for imgcnt, s in enumerate(cat.values()):
    s.compute(aa)
    print '%d / %d' % (imgcnt + 1, len(cat.values()))
    print 'Pointing (ra, dec):', s._ra, s._dec
    if abs(aa.lat - s._dec) > n.pi/2:
        print '    Source never rises: skipping'
        continue
    src = a.fit.SrcCatalog([s])

    def to_fits(t,i):
        filename = (opts.prefix+t+opts.fmt+'.fits') % imgcnt
        print 'Saving data to', filename
        while len(i.shape) < 4: i.shape = i.shape + (1,)
        cen = ephem.Equatorial(s.ra, s.dec, epoch=aa.epoch)
        cen = ephem.Equatorial(cen, epoch=ephem.J2000)
        L,M = im.get_LM()
        a.img.to_fits(filename, i, clobber=True,
            object=s.src_name, obs_date=str(aa.date),
            ra=cen.ra*a.img.rad2deg, dec=cen.dec*a.img.rad2deg, epoch=2000.,
            d_ra=L[-1,-1]*a.img.rad2deg, d_dec=M[-1,-1]*a.img.rad2deg,
            freq=n.average(aa.ants[0].beam.afreqs),
        )

    fdim = (opts.prefix+'dim'+opts.fmt+'.fits') % imgcnt
    fdbm = (opts.prefix+'dbm'+opts.fmt+'.fits') % imgcnt
    if os.path.exists(fdim) and os.path.exists(fdbm):
        print 'Found %s and %s, so skipping gridding/phasing.' % (fdim,fdbm)
        dim, kwds = a.img.from_fits(fdim)
        dbm, kwds = a.img.from_fits(fdbm)
        dim,dbm = dim.squeeze(), dbm.squeeze()
    else:
        # Gather data
        us,vs,ws = [],[],[]
        ds,wgts = [], []
        cnt, curtime = 0, None
        if opts.no_w: im = a.img.Img(opts.size, opts.res, mf_order=0)
        else: im = a.img.ImgW(opts.size, opts.res, mf_order=0)
        for L in locs:
          aa = aas[L]
          # Read each file
          for filename in locs[L]:
            sys.stdout.write('.'); sys.stdout.flush()
            uv = a.miriad.UV(filename)
            a.scripting.uv_selector(uv, opts.ant, opts.pol)
            # Read all data from each file
            for (crd,t,(i,j)),d,f in uv.all(raw=True):
                if curtime != t:
                    curtime = t
                    cnt = (cnt + 1) % opts.decimate
                    if cnt == 0:
                        aa.set_jultime(t)
                        src.compute(aa)
                        s_eq = src.get_crds('eq', ncrd=3)
                        aa.sim_cache(s_eq)
                if cnt != 0: continue
                d,f = d.take(chans), f.take(chans)
                if not opts.skip_amp: d /= aa.passband(i,j)
                try:
                    # Throws PointingError if not up:
                    d = aa.phs2src(d, s, i, j)
                    u,v,w = aa.gen_uvw(i,j,src=s)
                    if not opts.skip_bm:
                        # Calculate beam strength for weighting purposes
                        wgt = aa.bm_response(i,j,pol=opts.pol).squeeze()
                        # Optimal SNR: down-weight beam-attenuated data 
                        # by another factor of the beam response.
                        d *= wgt; wgt *= wgt
                    else: wgt = n.ones(d.shape, dtype=n.float)
                except(a.ant.PointingError): continue
                valid = n.logical_not(f)
                d = d.compress(valid)
                if len(d) == 0: continue
                ds.append(d)
                us.append(u.compress(valid))
                vs.append(v.compress(valid))
                ws.append(w.compress(valid))
                wgts.append(wgt.compress(valid))
                # If data buffer is full, grid data
                if len(ds) * len(chans) > opts.buf_thresh:
                    sys.stdout.write('|'); sys.stdout.flush()
                    ds = n.concatenate(ds)
                    us = n.concatenate(us)
                    vs = n.concatenate(vs)
                    ws = n.concatenate(ws)
                    wgts = n.concatenate(wgts).flatten()
                    # Grid data into UV matrix
                    (us,vs,ws),ds,wgts = im.append_hermitian((us,vs,ws),ds,wgts)
                    im.put((us,vs,ws), ds, wgts)
                    us,vs,ws = [],[],[]
                    ds,wgts = [],[]

        # Grid remaining data into UV matrix
        sys.stdout.write('|\n'); sys.stdout.flush()
        ds = n.concatenate(ds)
        us = n.concatenate(us)
        vs = n.concatenate(vs)
        ws = n.concatenate(ws)
        wgts = n.concatenate(wgts).flatten()
        if len(ds) == 0: raise ValueError('No data to use.')
        (us,vs,ws),ds,wgts = im.append_hermitian((us,vs,ws),ds,wgts)
        im.put((us,vs,ws), ds, wgts)
        if opts.uniform > 0: im.uniform_wgt(thresh=opts.uniform)

        # Form dirty images/beams
        uvs = a.img.recenter(n.abs(im.uv).astype(n.float), (DIM/2,DIM/2))
        bms = a.img.recenter(n.abs(im.bm[0]).astype(n.float), (DIM/2,DIM/2))
        dim = im.image((DIM/2, DIM/2))
        dbm = im.bm_image(term=0, center=(DIM/2,DIM/2))
        if 'uvs' in outputs: to_fits('uvs', uvs)
        if 'bms' in outputs: to_fits('bms', bms)
        if 'dim' in outputs: to_fits('dim', dim)
        if 'dbm' in outputs: to_fits('dbm', dbm)

    dbm = a.img.recenter(dbm, (DIM/2+1,DIM/2))
    bm_gain = a.img.beam_gain(dbm)
    print 'Gain of dirty beam:', bm_gain

    # Deconvolve
    if opts.deconv == 'mem':
        cim,info = a.deconv.maxent_findvar(dim, dbm, f_var0=opts.var,
            maxiter=opts.maxiter, verbose=True, tol=opts.tol, maxiterok=True)
    elif opts.deconv == 'lsq':
        cim,info = a.deconv.lsq(dim, dbm, 
            maxiter=opts.maxiter, verbose=True, tol=opts.tol)
    elif opts.deconv == 'cln':
        cim,info = a.deconv.clean(dim, dbm, 
            maxiter=opts.maxiter, verbose=True, tol=opts.tol)
    elif opts.deconv == 'ann':
        cim,info = a.deconv.anneal(dim, dbm, maxiter=opts.maxiter, 
            cooling=lambda i,x: opts.tol*(1-n.cos(i/50.))*(x**2), verbose=True)
    else:
        cim,info = n.zeros_like(dim), {'res':dim}

    rim = info['res'] / bm_gain
    bim = rim + cim
    if 'cim' in outputs: to_fits('cim', cim)
    if 'rim' in outputs: to_fits('rim', rim)
    if 'bim' in outputs: to_fits('bim', bim)




#! /usr/bin/env python
import aipy as a, numpy as n, time, os, sys

uvi = a.miriad.UV('test.uv')
freqs = n.arange(uvi['nchan'], dtype=n.float) * uvi['sdf'] + uvi['sfreq']
loc = ('0:00', '0:00')
beam = a.sim.BeamFlat(freqs)
ants = [
    a.sim.Antenna(0,0,0,beam),
    a.sim.Antenna(100,0,0,beam),
    a.sim.Antenna(0,100,0,beam),
    a.sim.Antenna(100,100,0,beam),
]
aa = a.sim.AntennaArray(ants=ants, location=loc)
src = a.sim.RadioFixedBody('0','0', 1e4)
if os.path.exists('test.uvs'):
    print 'test.uvs exists... skipping'
    sys.exit(0)
uvo = a.miriad.UV('test.uvs', status='new')
uvo.init_from_uv(uvi)

curtime = None
cat = a.sim.SrcCatalog([src] * 4)
def mfunc(uv, p, data):
    global curtime
    uvw,t,(i,j) = p
    if curtime != t:
        curtime = t
        aa.set_jultime(t)
        src.compute(aa)
        s_eqs = cat.get_crds('eq')
        fluxes = cat.get_fluxes()
        indices = cat.get_indices()
        mfreqs = cat.get_mfreqs()
        aa.sim_cache(s_eqs=s_eqs,fluxes=fluxes,indices=indices,mfreqs=mfreqs)
    d = aa.sim(i, j)
    return p, n.ma.array(d, mask=data.mask)

t0 = time.time()
uvo.pipe(uvi, mfunc=mfunc)
print 'Simulation finished in %f seconds.' % (time.time() - t0)
del(uvo)
print 'Checking against knowngood.uvs ...'
uv1 = a.miriad.UV('knowngood.uvs')
uv2 = a.miriad.UV('test.uvs')

for (p1,d1),(p2,d2) in zip(uv1.all(), uv2.all()):
    dif = (d1.filled(0) - d2.filled(0)).sum()
    if dif > 1e-10:
        raise ValueError('Failed verification (%f > 1e-10).' % dif)
print 'Passed.'



#!/usr/bin/python
"""
Removes crosstalk from PAPER data by filtering delays and fringes that 
are unlikely to correspond to celestial sources.  Has shortcoming that
flagged data cause power to be scattered by filtering process.

Author: Aaron Parsons
"""
import os, sys, aipy as a, numpy as n

def gen_skypass_delay(aa, sdf, nchan, max_bl_frac=1.):
    bin_dly = 1. / (sdf * nchan)
    filters = {}
    for i,j in [aa.bl2ij(bl) for bl in aa.bl_order]:
        bl = aa.ij2bl(i,j)
        max_bl = aa.get_baseline(i,j)
        max_bl = max_bl_frac * n.sqrt(n.dot(max_bl, max_bl))
        dly = aa.get_delay(i,j)
        uthresh, lthresh = (dly + max_bl)/bin_dly + 1, (dly - max_bl)/bin_dly
        uthresh, lthresh = int(n.round(uthresh)), int(n.round(lthresh))
        f = n.ones((nchan,), dtype=n.float)
        f[uthresh:lthresh] = 0
        filters[bl] = f
    return filters

def gen_skypass_fringe(aa, inttime, N_int, nchan, margin=1.2):
    freqs = aa.ants[0].beam.freqs
    bin_dly = 1. / (inttime * N_int)
    dth_dt = 2*n.pi / a.const.sidereal_day
    filters = {}
    for i,j in [aa.bl2ij(bl) for bl in aa.bl_order]:
        bl = aa.ij2bl(i,j)
        max_bl = aa.get_baseline(i,j)[:2]
        max_bl = n.sqrt(n.dot(max_bl, max_bl)) * freqs
        max_fr = margin * dth_dt * max_bl
        uthresh, lthresh = max_fr / bin_dly + 1, -max_fr / bin_dly
        uthresh = n.round(uthresh).astype(n.int)
        lthresh = n.round(lthresh).astype(n.int)
        f = n.ones((N_int, nchan), dtype=n.float)
        for c in range(nchan):
            f[uthresh[c]:lthresh[c],c] = 0
        filters[bl] = f
    return filters
    
if __name__ == '__main__':
    import optparse

    o = optparse.OptionParser()
    o.set_usage('xtalk2.py [options] *.uv')
    o.set_description(__doc__)
    a.scripting.add_standard_options(o, loc=True)
    o.add_option('--clean', dest='clean', type='float', default=1e-3,
        help='Deconvolve delay-domain data by the "beam response" that results from flagged data.  Specify a tolerance for termination (usually 1e-2 or 1e-3).')
    opts,args = o.parse_args(sys.argv[1:])

    # Parse command-line options
    uv = a.miriad.UV(args[0])
    nchan, inttime = uv['nchan'], uv['inttime']
    sdf, sfreq = uv['sdf'], uv['sfreq']
    aa = a.loc.get_aa(opts.loc, sdf, sfreq, nchan)
    skypass_delay = gen_skypass_delay(aa, sdf, nchan, max_bl_frac=1.)
    del(uv)

    for uvfile in args:
        print 'Working on', uvfile
        uvofile = uvfile+'X'
        if os.path.exists(uvofile):
            print uvofile, 'exists, skipping.'
            continue
        uvi = a.miriad.UV(uvfile)

        # Gather all data from file
        data = {}
        print '    Performing delay transform and cleaning...'
        for (uvw,t,(i,j)),d,f in uvi.all(raw=True):
            bl = aa.ij2bl(i,j)
            if i == j: continue
            # Filter out delays that don't correspond to the sky
            flags = n.logical_not(f).astype(n.float)
            gain = n.sqrt(n.average(flags**2))
            ker = n.fft.ifft(flags)
            d = n.where(f, 0, d)
            d = n.fft.ifft(d)
            if not n.all(d == 0):
                d, info = a.deconv.clean1d(d, ker, tol=opts.clean)
                d += info['res'] / gain
            d *= skypass_delay[bl]
            d = n.fft.fft(d)
            try: data[bl].append(d)
            except(KeyError): data[bl] = [d]

        print '    Performing fringe rate transform and cleaning...'
        # Filter out fringes that don't correspond to the sky
        ntimes = len(data.values()[0])
        skypass_fringe = gen_skypass_fringe(aa, inttime, ntimes, nchan)
        for bl in data:
            d = n.array(data[bl])
            flags = n.where(d[:,0] != 0, 1., 0.)
            gain = n.sqrt(n.average(flags**2))
            ker = n.fft.ifft(flags)
            d = n.fft.ifft(d, axis=0)
            for chan in range(d.shape[1]):
                d[:,chan],info = a.deconv.clean1d(d[:,chan],ker,tol=opts.clean)
                d[:,chan] += info['res'] / gain
            d *= skypass_fringe[bl]
            # Also kill 0 fringe (narrows FOV, but kills a lot of xtalk and rfi)
            d[0] = 0
            d = n.fft.fft(d, axis=0)
            data[bl] = d

        cnt = {}
        for bl in data: cnt[bl] = 0

        # Generate a pipe for swapping in new data
        def mfunc(uv, p, d, f):
            uvw, t, (i,j) = p
            if i == j: return p,d,f
            bl = a.miriad.ij2bl(i,j)
            d = data[bl][cnt[bl],:]
            cnt[bl] += 1
            return p, d, f

        # Apply the pipe to the data
        uvi.rewind()
        uvo = a.miriad.UV(uvofile, status='new')
        uvo.init_from_uv(uvi)
        uvo.pipe(uvi, mfunc=mfunc, raw=True,
            append2hist='XTALK2: clean=%f\n' % (opts.clean))



#! /usr/bin/env python
"""
Removes crosstalk from PAPER data by modeling it over the course of a day
as a static per-channel cross-coupling that rises and falls uniformly across
the band with changing input amplitude.  Steps for crosstalk removal are:
run "xtalk3.py -o" on 1 day of UV files (which should be flagged for RFI, and
if possible, have a sky model removed) to 
generate *.xtalk files.  Then run "xtalk3.py -r" on same UV files to 
reprocess *.xtalk files, separating them into static shape/uniform gain 
terms that are stored in *.xtalk.rep files.  Finally, run "xtalk3.py -i" on 
UV files from the same JD (but which need not have RFI flagged or a sky model
removed) to use the *.xtalk.rep model to remove crosstalk from the visibility 
data.

Author: Aaron Parsons
"""

import aipy as a, numpy as n, sys, os, optparse, pickle

o = optparse.OptionParser()
o.set_usage('xtalk3.py [options] *.uv')
o.set_description(__doc__)
o.add_option('-i', '--infile', dest='infile', action='store_true',
    help='Apply xtalk calibrations generated with the -o option.')
o.add_option('-o', '--outfile', dest='outfile', action='store_true',
    help='Rather than apply the calibrations to the data, store them in a file (named by JD) to apply to a different file with the same JD.')
o.add_option('-r', '--reprocess', dest='reprocess', action='store_true',
    help='Reprocess xtalk files for the specified UV files to generate an average crosstalk profile and a time-dependent gain for scaling that profile.  The results will be saved to *.xtalk.rep files which will have precedence over *.xtalk files.')
o.add_option('-c', '--chan', dest='chan', default='160_720',
    help='Channels range to use for normalization when reprocessing.')
opts,args = o.parse_args(sys.argv[1:])

assert(not opts.infile or not opts.outfile)

if opts.reprocess:
    chans = map(int, opts.chan.split('_'))
    xsum, cnt, gain, times = {}, {}, {}, []
    for filename in args:
        uv = a.miriad.UV(filename)
        (uvw,jd,(i,j)),d,f = uv.read(raw=True)
        xfile = '%f.xtalk' % jd
        print xfile
        times.append(jd)
        if not os.path.exists(xfile):
            print xfile, 'does not exist.  Skipping...'
            continue
        f = open(xfile); xtalk = pickle.load(f); f.close()
        for bl in xtalk:
            dat = n.array(xtalk[bl])
            adat = n.ma.masked_equal(n.abs(dat), 0)
            if not gain.has_key(bl): gain[bl] = []
            gain[bl].append(n.average(adat[chans[0]:chans[1]]))
            dat /= gain[bl][-1]
            xsum[bl] = xsum.get(bl, 0) + dat
            cnt[bl] = cnt.get(bl, 0) + n.logical_not(adat.mask).astype(n.int)
    for bl in xsum: xsum[bl] /= n.where(cnt[bl] == 0, 1, cnt[bl])
    for c, jd in enumerate(times):
        repfile = '%f.xtalk.rep' % jd
        for bl in xsum: xtalk[bl] = list(gain[bl][c] * xsum[bl])
        print 'Writing', repfile
        f = open(repfile, 'w')
        pickle.dump(xtalk, f)
        f.close()
    import sys; sys.exit(0)

guess, cnt, xtalk = {}, {}, {}
for filename in args:
    print filename
    if not opts.outfile and os.path.exists(filename+'x'):
        print filename+'x', 'exists.  Skipping...'
        continue
    uv = a.miriad.UV(filename)
    uv.select('auto',0, 0, include=False)
    (uvw,jd,(i,j)),d,f = uv.read(raw=True)
    uv.rewind()
    if opts.infile:
        xfile = '%f.xtalk.rep' % jd
        if not os.path.exists(xfile): xfile = '%f.xtalk' % jd
        if not os.path.exists(xfile):
            print xfile, 'does not exist.  Skipping...'
            continue
        print '    using', xfile
        f = open(xfile); xtalk = pickle.load(f); f.close()
        for bl in xtalk: xtalk[bl] = n.array(xtalk[bl])
    else:
        guess, cnt, xtalk = {}, {}, {}
        for (uvw,t,(i,j)),d,f in uv.all(raw=True):
            bl = a.miriad.ij2bl(i,j)
            if not guess.has_key(bl): guess[bl], cnt[bl] = 0, 0
            guess[bl] += n.where(f, 0, d)
            cnt[bl] += n.logical_not(f)
        del(uv)
        for bl in guess: xtalk[bl] = guess[bl] / n.clip(cnt[bl], 1, n.Inf)
    if opts.outfile:
        for bl in xtalk: xtalk[bl] = list(xtalk[bl])
        print 'Writing %f.xtalk' % jd
        f = open('%f.xtalk' % jd, 'w')
        pickle.dump(xtalk, f)
        f.close()
    else:
        def mfunc(uv, p, d, f):
            uvw,t,(i,j) = p
            bl = a.miriad.ij2bl(i,j)
            if xtalk.has_key(bl): return p, d - xtalk[bl], f
            else: return p, d, f
        uvi = a.miriad.UV(filename)
        uvo = a.miriad.UV(filename+'x', status='new')
        uvo.init_from_uv(uvi)
        uvo.pipe(uvi, mfunc=mfunc,
            append2hist='XTALK3: removed crosstalk\n', raw=True)



"""
A package for interfacing to Miriad: the Multichannel Image Reconstruction
Image Analysis and Display package for reducing interferometric data for
radio telescopes.
"""

__version__ = '0.1.1'

import numpy as n, _miriad

def echo(uv, p, d): return p, d

str2pol = {
    'I' :  1,   # Stokes Paremeters
    'Q' :  2,
    'U' :  3,
    'V' :  4,
    'rr': -1,   # Circular Polarizations
    'll': -2,
    'rl': -3,
    'lr': -4,
    'xx': -5,   # Linear Polarizations
    'yy': -6,
    'xy': -7,
    'yx': -8,
}

pol2str = {}
for k in str2pol: pol2str[str2pol[k]] = k

itemtable = {
    'obstype' : 'a',
    'history' : 'a',
    'vartable': 'a',
    #'visdata' : '?',
    #'flags'   : 'i',
    #'wflags'  : 'i',
    #'gains'   : '?',
    'ngains'  : 'i',
    'nfeeds'  : 'i',
    'ntau'    : 'i',
    'nsols'   : 'i',
    'interval': 'd',
    'leakage' : 'c',
    'freq0'   : 'd',
    'freqs'   : '?',
    'bandpass': 'c',
    'nspect0' : 'i',
    'nchan0'  : 'i',
}

data_types = {
    'a' : "ascii (NULL terminated)",
    'r' : "real (32 bin IEEE)",
    'd' : "double (64 bit)",
    'c' : "complex (2 * 32 bit IEEE)",
    'i' : "integer (32 bit two's complement)",
}

#  _   ___     __
# | | | \ \   / /
# | | | |\ \ / / 
# | |_| | \ V /  
#  \___/   \_/   

class UV(_miriad.UV):
    """Top-level interface to a Miriad UV data set."""
    def __init__(self, filename, status='old'):
        """Open a miriad file.  status can be ['old','new','append']."""
        assert(status in ['old', 'new', 'append'])
        _miriad.UV.__init__(self, filename, status)
        self.status = status
        self.nchan = 4096
        if status == 'old':
            self.vartable = self._gen_vartable()
            self.read(); self.rewind() # Update variables for the user
            try: self.nchan = self['nchan']
            except(KeyError): pass
        else: self.vartable = {}
    def _gen_vartable(self):
        """Generate table of variables and types from the vartable header."""
        vartable = {}
        for line in self._rdhd('vartable').split('\n'):
            try:
                type, name = line.split()
                vartable[name] = type
            except(ValueError): pass
        return vartable
    def vars(self):
        """Return a list of available variables."""
        return self.vartable.keys()
    def items(self):
        """Return a list of available header items."""
        items = []
        for i in itemtable:
            try:
                _miriad.hdaccess(self.haccess(i, 'read'))
                items.append(i)
            except(IOError): pass
        return items
    def _rdhd(self, name):
        """Provide read access to header items via low-level calls."""
        itype = itemtable[name]
        if itype == '?': return self._rdhd_special(name)
        h = self.haccess(name, 'read')
        rv = []
        if len(itype) == 1:
            if itype == 'a': offset = 0
            else:
                t, offset = _miriad.hread_init(h)
                assert(itype == t)
            while True:
                try: c, o = _miriad.hread(h, offset, itype)
                except(IOError): break
                if itype == 'a': c = c[:o]
                rv.append(c)
                offset += o
            if itype == 'a': rv = ''.join(rv)
        else:
            t, offset = _miriad.hread_init(h); assert(t == 'b')
            for t in itype:
                v, o = _miread.hread(h, offset, t)
                rv.append(v); offset += o
        _miriad.hdaccess(h)
        if len(rv) == 1: return rv[0]
        elif type(rv) == str: return rv
        else: return n.array(rv)
    def _wrhd(self, name, val):
        """Provide write access to header items via low-level calls."""
        type = itemtable[name]
        if type == '?': return self._wrhd_special(name, val)
        h = self.haccess(name, 'write')
        if len(type) == 1:
            try: len(val)
            except(TypeError): val = [val]
            if type == 'a': offset = 0
            else: offset = _miriad.hwrite_init(h, type)
            for v in val: offset += _miriad.hwrite(h, offset, v, type)
        else:
            offset = _miriad.hwrite_init(h, 'b')
            for v, t in zip(val,type): offset += _miriad.hwrite(h,offset,v,t)
        _miriad.hdaccess(h)
    def _rdhd_special(self, name):
        """Provide read access to special header items of type '?' to _rdhd"""
        if name == 'freqs':
            h = self.haccess(name, 'read')
            c, o = _miriad.hread(h, 0, 'i')
            rv = [c]; offset = 8
            while True:
                try:
                    c, o = _miriad.hread(h, offset, 'i')
                    rv.append(c); offset += 8
                    c, o = _miriad.hread(h, offset, 'd')
                    rv.append(c); offset += 8
                    c, o = _miriad.hread(h, offset, 'd')
                    rv.append(c); offset += 8
                except(IOError): break
            _miriad.hdaccess(h)
            return rv
        else: raise ValueError('Unknown special header: ' + name)
    def _wrhd_special(self, name, val):
        """Provide write access to special header items of type '?' to _rdhd"""
        if name == 'freqs':
            h = self.haccess(name, 'write')
            o = _miriad.hwrite(h, 0, val[0], 'i')
            offset = 8
            for i,v in enumerate(val[1:]):
                if i % 3 == 0: o = _miriad.hwrite(h, offset, v, 'i')
                else: o = _miriad.hwrite(h, offset, v, 'd')
                offset += 8
            _miriad.hdaccess(h)
        else: raise ValueError('Unknown special header: ' + name)
    def __getitem__(self, name):
        """Allow access to variables and header items via uv[name]."""
        try:
            type = self.vartable[name]
            return self._rdvr(name, type)
        except(KeyError):
            type = itemtable[name]
            return self._rdhd(name)
    def __setitem__(self, name, val):
        """Allow setting variables and header items via uv[name] = val."""
        try:
            type = self.vartable[name]
            self._wrvr(name,type,val)
        except(KeyError):
            self._wrhd(name,val)
    def select(self, name, n1, n2, include=1):
        """Choose which data are returned by read().  
            name    This can be one of 'time','antennae','visibility',
                    'uvrange','pointing','amplitude','window','or','dra',
                    'ddec','uvnrange','increment','ra','dec','and', 'clear',
                    'on','polarization','shadow','auto','dazim','delev'
            n1,n2   Generally this is the range of values to select. For
                    'antennae', this is the two antennae pair to select
                    (indexed from 0); a -1 indicates 'all antennae'.
                    For 'shadow', a zero indicates use 'antdiam' variable.
                    For 'on','window','polarization','increment','shadow' only
                    p1 is used.
                    For 'and','or','clear','auto' p1 and p2 are ignored.
            include If true, the data is selected. If false, the data is
                    discarded. Ignored for 'and','or','clear'."""
        if name == 'antennae':
            n1 += 1; n2 += 1
        self._select(name, float(n1), float(n2), int(include))
    def read(self, raw=False):
        """Return the next data record.  Calling this function causes 
        vars to change to reflect the record which this function returns.
        'raw' causes data and flags to be returned seperately."""
        preamble, data, flags, nread = self.raw_read(self.nchan)
        if nread == 0: raise IOError("No data read")
        flags = n.logical_not(flags)
        if raw: return preamble, data, flags
        return preamble, n.ma.array(data, mask=flags)
    def all(self, raw=False):
        """Provide an iterator over preamble, data.  Allows constructs
        like: for preamble, data in uv.all(): ..."""
        while True:
            try: yield self.read(raw=raw)
            except(IOError): return
    def write(self, preamble, data, flags=None):
        """Write the next data record.  data must be a complex, masked
        array.  preamble must be (uvw, t, (i,j)), where uvw is an array of 
        u,v,w, t is the Julian date, and (i,j) is an antenna pair."""
        if data is None: return
        if not flags is None: flags = n.logical_not(flags)
        elif len(data.mask.shape) == 0:
            flags = n.ones(data.shape)
            data = data.unmask()
        else:
            flags = n.logical_not(data.mask)
            #data = data.filled(0)
            data = data.data
        self.raw_write(preamble,data.astype(n.complex64),flags.astype(n.int32))
    def init_from_uv(self, uv, override={}, exclude=[]):
        """Initialize header items and variables from another UV.  Those in 
        override will be overwritten by override[k], and tracking will be 
        turned off (meaning they will not be updated in pipe()).  Those in
        exclude are omitted completely."""
        for k in uv.items():
            if k in exclude: continue
            elif k in override: self._wrhd(k, override[k])
            else: self._wrhd(k, uv[k])
        self.vartable = {}
        for k in uv.vars():
            if k in exclude: continue
            # I don't understand why reading 'corr' segfaults miriad,
            # but it does.  This is a cludgy work-around.
            elif k == 'corr': continue
            elif k in override:
                self.vartable[k] = uv.vartable[k]
                self._wrvr(k, uv.vartable[k], override[k])
            else:
                self.vartable[k] = uv.vartable[k]
                self._wrvr(k, uv.vartable[k], uv[k])
                uv.trackvr(k, 'c') # Set to copy when copyvr() called 
    def pipe(self, uv, mfunc=echo, append2hist='', raw=False):
        """Pipe in data from another UV through the function
        mfunc(uv,preamble,data), which should return (preamble,data).  If 
        mfunc is not provided, the dataset will just be cloned, and if the 
        returned data is None, it will be omitted.  The string 'append2hist' 
        will be appended to history."""
        self._wrhd('history', self['history'] + append2hist)
        # Pipe all data through mfunc
        if raw:
            for p,d,f in uv.all(raw=raw):
                np, nd, nf = mfunc(uv, p, d, f)
                self.copyvr(uv)
                self.write(np, nd, nf)
        else:
            for p, d in uv.all():
                np, nd = mfunc(uv, p, d)
                self.copyvr(uv)
                self.write(np, nd)
    def add_var(self, name, type):
        """Add a variable of the specified type to a UV file."""
        self.vartable[name] = type

def bl2ij(bl):
    bl = int(bl)
    return (bl>>8)-1, (bl&255) - 1

def ij2bl(i, j):
    if i > j: i,j = j,i
    return (i+1)<<8 | (j+1)




#! /usr/bin/env python
"""
This is a general-purpose script for making faceted, spherical maps (stored in
Healpix FITS format) from individual "flat" maps stored in FITS files.

Author: Aaron Parsons
"""

import sys, numpy as n, os, aipy as a, optparse

o = optparse.OptionParser()
o.set_usage('mk_map.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, pol=True, chan=True)
o.add_option('-i', '--interpolate', dest='interpolate', action='store_true',
    help='Use sub-pixel interpolation when gridding data to healpix map.')
o.add_option('-m', '--map', dest='map',
    help='The skymap file to use.  If it exists, new data will be added to the map.  Othewise, the file will be created.')
o.add_option('--nside', dest='nside', type='int', default=256,
    help='NSIDE parameter for map, if creating a new file.')
o.add_option('--fwidth', dest='fwidth', type='float', default=10,
    help='Width of gaussian, in degrees, used to downweight data away from pointing center for each facet.  Default 10.')
opts, args = o.parse_args(sys.argv[1:])

# Open skymap
if os.path.exists(opts.map): skymap = a.map.Map(fromfits=opts.map)
else: skymap = a.map.Map(nside=opts.nside)
skymap.set_interpol(opts.interpolate)

for i, filename in enumerate(args):
    img, kwds = a.img.from_fits(filename)
    img = img.squeeze()
    s = a.ant.RadioFixedBody(kwds['ra']*a.img.deg2rad,
                             kwds['dec']*a.img.deg2rad)
    DIM = img.shape[0]
    RES = 1. / (kwds['d_ra'] * a.img.deg2rad * DIM)
    im = a.img.Img(DIM*RES, RES, mf_order=0)
    print '-----------------------------------------------------------'
    print 'Reading file %s (%d / %d)' % (filename, i + 1, len(args))
    print kwds
    print 'Pointing (ra, dec):', s._ra, s._dec
    # Get coordinates of image pixels in original (J2000) epoch
    ex,ey,ez = im.get_eq(s._ra, s._dec, center=(DIM/2,DIM/2))
    tx,ty,tz = im.get_top(center=(DIM/2,DIM/2))
    # Define a weighting for gridding data into the skymap
    map_wgts = n.exp(-(tx**2 + ty**2) / n.sin(opts.fwidth * a.img.deg2rad)**2)
    map_wgts.shape = (map_wgts.size,)
    valid = n.logical_not(map_wgts.mask)
    ex = ex.compress(valid); ey = ey.compress(valid); ez = ez.compress(valid)
    map_wgts = map_wgts.compress(valid)
    img = img.flatten()
    img = img.compress(valid)
    # Put the data into the skymap
    skymap.add((ex,ey,ez), map_wgts, img)
skymap.to_fits(opts.map, clobber=True)




"""
Module for representing antenna array geometry and for generating
phasing information.
"""
import ephem, math, numpy as n, coord, const, _cephes

class PointingError(Exception):
    """An error to throw if a source is below the horizon."""
    def __init__(self, value): self.parameter = value
    def __str__(self): return repr(self.parameter)

#  _   _ _   _ _ _ _           _____                 _   _                 
# | | | | |_(_) (_) |_ _   _  |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
# | | | | __| | | | __| | | | | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
# | |_| | |_| | | | |_| |_| | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
#  \___/ \__|_|_|_|\__|\__, | |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
#                      |___/        

def juldate2ephem(num):
    """Convert Julian date to ephem date, measured from noon, Jan. 1, 1900."""
    return ephem.date(num - 2415020)

def ephem2juldate(num):
    """Convert ephem date (measured from noon, Jan. 1, 1900) to Julian date."""
    return float(num + 2415020)

#  ____           _ _       ____            _       
# |  _ \ __ _  __| (_) ___ | __ )  ___   __| |_   _ 
# | |_) / _` |/ _` | |/ _ \|  _ \ / _ \ / _` | | | |
# |  _ < (_| | (_| | | (_) | |_) | (_) | (_| | |_| |
# |_| \_\__,_|\__,_|_|\___/|____/ \___/ \__,_|\__, |
#                                             |___/ 

class RadioBody:
    """The location of a celestial source."""
    def __init__(self, name='', **kwargs):
        self.src_name = name
    def compute(self, observer):
        """Update coordinates relative to the provided observer.  Must be
        called at each time step before accessing information."""
        # Generate a map for projecting baselines to uvw coordinates
        self.map = coord.eq2top_m(observer.sidereal_time()-self.ra, self.dec)
        # Generate vector pointing at src in various coordinates
        self.e_vec = coord.radec2eq((self.ra, self.dec))
        self.t_vec = coord.azalt2top((self.az, self.alt))
    def get_crd(self, crdsys, ncrd=3):
        """Return the coordinates of this location in the desired coordinate
        system ('eq','top') in the current epoch.  If ncrd=2, angular
        coordinates (ra/dec or az/alt) are returned, and if ncrd=3,
        xyz coordinates are returned."""
        assert(crdsys in ('eq','top'))
        assert(ncrd in (2,3))
        if crdsys == 'eq':
            if ncrd == 2: return (self.ra, self.dec)
            return self.e_vec
        else:
            if ncrd == 2: return (self.az, self.alt)
            return self.t_vec

#  ____           _ _       _____ _              _ ____            _       
# |  _ \ __ _  __| (_) ___ |  ___(_)_  _____  __| | __ )  ___   __| |_   _ 
# | |_) / _` |/ _` | |/ _ \| |_  | \ \/ / _ \/ _` |  _ \ / _ \ / _` | | | |
# |  _ < (_| | (_| | | (_) |  _| | |>  <  __/ (_| | |_) | (_) | (_| | |_| |
# |_| \_\__,_|\__,_|_|\___/|_|   |_/_/\_\___|\__,_|____/ \___/ \__,_|\__, |
#                                                                    |___/ 

class RadioFixedBody(ephem.FixedBody, RadioBody):
    """A source at fixed RA,DEC.  Combines ephem.FixedBody with RadioBody."""
    def __init__(self, ra, dec, name='', epoch=ephem.J2000, **kwargs):
        RadioBody.__init__(self, name=name)
        ephem.FixedBody.__init__(self)
        self._ra, self._dec = ra, dec
        self._epoch = epoch
    def compute(self, observer):
        ephem.FixedBody.compute(self, observer)
        RadioBody.compute(self, observer)

#  ____           _ _      ____                  _       _ 
# |  _ \ __ _  __| (_) ___/ ___| _ __   ___  ___(_) __ _| |
# | |_) / _` |/ _` | |/ _ \___ \| '_ \ / _ \/ __| |/ _` | |
# |  _ < (_| | (_| | | (_) |__) | |_) |  __/ (__| | (_| | |
# |_| \_\__,_|\__,_|_|\___/____/| .__/ \___|\___|_|\__,_|_|
#                               |_|                        

class RadioSpecial(RadioBody, object):
    """A moving source (Sun,Moon,planets).  Combines ephem versions of these 
    objects with RadioBody."""
    def __init__(self, name, **kwargs):
        """`name' is used to lookup appropriate ephem celestial object."""
        RadioBody.__init__(self, name=name)
        self.Body = eval('ephem.%s()' % name)
    def __getattr__(self, nm):
        """First try to access attribute from this class, but if that fails, 
        try to get it from the underlying ephem object."""
        try: return object.__getattr__(self, nm)
        except(AttributeError): return self.Body.__getattribute__(nm)
    def __setattr__(self, nm, val):
        """First try to set attribute for this class, buf if that fails, 
        try to set it for the underlying ephem object."""
        try: object.__setattr__(self, nm, val)
        except(AttributeError): return setattr(self.Body, nm, val)
    def compute(self, observer):
        self.Body.compute(observer)
        RadioBody.compute(self, observer)

#  ____            ____      _        _             
# / ___| _ __ ___ / ___|__ _| |_ __ _| | ___   __ _ 
# \___ \| '__/ __| |   / _` | __/ _` | |/ _ \ / _` |
#  ___) | | | (__| |__| (_| | || (_| | | (_) | (_| |
# |____/|_|  \___|\____\__,_|\__\__,_|_|\___/ \__, |
#                                             |___/ 

class SrcCatalog(dict):
    """A catalog of celestial sources."""
    def __init__(self, srcs, **kwargs):
        dict.__init__(self)
        for s in srcs: self.add_src(s)
    def add_src(self, src):
        """Add provided src object (RadioFixedBody,RadioSpecial) to catalog."""
        self[src.src_name] = src
    def get_srcs(self, *args):
        """Return list of all src objects in catalog."""
        return [self[s] for s in args]
    def compute(self, observer):
        """Call compute method of all objects in catalog."""
        for s in self: self[s].compute(observer)
    def get_crds(self, crdsys, ncrd=3, srcs=None):
        """Return coordinates of all objects in catalog."""
        if srcs is None: srcs = self.keys()
        crds = n.array([self[s].get_crd(crdsys, ncrd=ncrd) for s in srcs])
        return crds.transpose()

#  ____
# | __ )  ___  __ _ _ __ ___
# |  _ \ / _ \/ _` | '_ ` _ \
# | |_) |  __/ (_| | | | | | |
# |____/ \___|\__,_|_| |_| |_|

class Beam:
    """Template for representing antenna beam pattern.  Beams also hold
    info about which frequencies are active (i.e. Antennas and AntennaArrays
    access frequencies through Beam)."""
    def __init__(self, freqs, active_chans=None, **kwargs):
        """freqs = frequencies (in GHz) at bin centers across spectrum
        active_chans = indices of channels to use for future calculations."""
        self.freqs = freqs
        self.select_chans(active_chans)
    def select_chans(self, active_chans=None):
        """Select only enumerated channels to use for future calculations."""
        if active_chans is None: active_chans = n.arange(self.freqs.size)
        self.chans = active_chans
        self.afreqs = self.freqs.take(active_chans)

#     _          _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` |
#  / ___ \| | | | ||  __/ | | | | | | (_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_|

class Antenna:
    """Representation of physical attributes of individual antenna."""
    def __init__(self, x, y, z, beam, delay=0., offset=0., **kwargs):
        """x,y,z = antenna coordinates in equatorial (ns) coordinates
        beam = Beam object
        delay = signal delay (linear phase vs. frequency)
        offset = signal phase offset (constant phase vs. frequency)"""
        self.pos = n.array((x,y,z), n.float64) # must be float64 for mir
        self.beam = beam
        self.delay = delay
        self._offset = offset
        try:
            len(offset)
            self.offset = n.polyval(offset, self.beam.afreqs)
        except(AttributeError,TypeError): self.offset = (offset % 1)
    def select_chans(self, active_chans=None):
        """Select only the specified channels for use in future calculations."""
        self.beam.select_chans(active_chans)
        try:
            len(self._offset)
            self.offset = n.polyval(self._offset, self.beam.afreqs)
        except(AttributeError,TypeError): self.offset = (self._offset % 1)
    def __tuple__(self): return (self.pos[0], self.pos[1], self.pos[2])
    def __list__(self): return [self.pos[0], self.pos[1], self.pos[2]]
    def __add__(self, a): return self.pos + a.pos
    __radd__ = __add__
    def __neg__(self): return -self.pos
    def __sub__(self, a): return self.pos - a.pos
    def __rsub__(self, a): return a.pos - self.pos

#     _                         _                    _   _             
#    / \   _ __ _ __ __ _ _   _| |    ___   ___ __ _| |_(_) ___  _ __  
#   / _ \ | '__| '__/ _` | | | | |   / _ \ / __/ _` | __| |/ _ \| '_ \ 
#  / ___ \| |  | | | (_| | |_| | |__| (_) | (_| (_| | |_| | (_) | | | |
# /_/   \_\_|  |_|  \__,_|\__, |_____\___/ \___\__,_|\__|_|\___/|_| |_|
#                         |___/                                        

class ArrayLocation(ephem.Observer):
    """The location and time of an observation."""
    def __init__(self, location):
        """location = (lat,long,[elev]) of array"""
        ephem.Observer.__init__(self)
        self.pressure = 0
        self.update_location(location)
    def update_location(self, location):
        """Initialize antenna array wth provided location.  May be (lat,long) 
        or (lat,long,elev)."""
        if len(location) == 2: self.lat, self.long = location
        else: self.lat, self.long, self.elev = location
    def set_jultime(self, t=None):
        """Set current time as a Julian date."""
        if t is None: t = ephem.julian_date()
        self.set_ephemtime(juldate2ephem(t))
    def set_ephemtime(self, t=None):
        """Set current time as derived from the ephem package."""
        if t is None: t = ephem.now()
        self.date, self.epoch = t, t

#     _          _                            _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _   / \   _ __ _ __ __ _ _   _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` | / _ \ | '__| '__/ _` | | | |
#  / ___ \| | | | ||  __/ | | | | | | (_| |/ ___ \| |  | | | (_| | |_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_/_/   \_\_|  |_|  \__,_|\__, |
#                                                                 |___/ 

class AntennaArray(ArrayLocation):
    """A collection of antennas, their spacings, and location/time of 
    observations."""
    def __init__(self, location, ants, **kwargs):
        """ location = (lat,long,[elev]) of array
        ants = list of Antenna objects."""
        ArrayLocation.__init__(self, location=location)
        self.update_antennas(ants)
        self.select_chans()
    def update_antennas(self, ants):
        """Update antenna array to use the provided list of Antenna objects."""
        self.ants = ants
        self.update()
    def update(self):
        """Update variables derived from antenna parameters/active channels."""
        bls,dlys,offs = [],[],[]
        self.bl_order = {}
        # Calculate deltas between antenna pairs
        for i, ai in enumerate(self.ants):
            for j, aj in enumerate(self.ants[i:]):
                bl = self.ij2bl(i, j+i)
                bls.append(aj - ai)
                dlys.append(aj.delay - ai.delay)
                offs.append(aj.offset - ai.offset)
                self.bl_order[bl] = len(bls) - 1
        self.bls,self.dlys,self.offs = n.array(bls),n.array(dlys),n.array(offs)
        # Compute (static) zenith baselines
        m = coord.eq2top_m(0., self.lat)
        self.zbls = n.dot(m, self.bls.transpose()).transpose()
    def set_ephemtime(self, t=None):
        """Set current time as derived from the ephem package.  Recalculates
        matrix for projecting baselines into current positions."""
        ArrayLocation.set_ephemtime(self, t=t)
        # Rotate baselines in eq coords to current sidereal time
        m = coord.rot_m(-self.sidereal_time(), n.array([0.,0.,1.]))
        self.ebls = n.dot(m, self.bls.transpose()).transpose()
    def select_chans(self, active_chans=None):
        """Select which channels are used in computations.  Default is all."""
        for a in self.ants: a.select_chans(active_chans)
        self.update()
    def ij2bl(self, i, j):
        """Convert baseline i,j (0 indexed) to Miriad's (i+1) << 8 | (j+1) 
        indexing scheme."""
        return (int(i)+1) << 8 | (int(j)+1)
    def bl2ij(self, bl):
        """Convert Miriad's (i+1) << 8 | (j+1) baseline indexing scheme to 
        i,j (0 indexed)"""
        bl = int(bl)
        return ((bl >> 8) & 255) - 1, (bl & 255) - 1
    def get_baseline(self, i, j, src='z'):
        """Return the baseline corresponding to i,j in various coordinate 
        projections: src='e' for current equatorial, 'z' for zenith 
        topocentric, 'r' for unrotated equatorial, or a RadioBody for
        projection toward that source."""
        b = self.bl_order[self.ij2bl(i,j)]
        if type(src) == str:
            if src == 'e': return self.ebls[b]
            elif src == 'z': return self.zbls[b]
            elif src == 'r': return self.bls[b]
            else: raise ValueError('Unrecognized source:' + src)
        try:
            if src.alt < 0:
                raise PointingError('%s below horizon' % src.src_name)
            map = src.map
        except(AttributeError):
            ra,dec = coord.eq2radec(src)
            map = coord.eq2top_m(self.sidereal_time() - ra, dec)
        return n.dot(map, self.bls[b]).transpose()
    def get_delay(self, i, j):
        """Return the delay corresponding to baseline i,j."""
        return self.dlys[self.bl_order[self.ij2bl(i,j)]]
    def get_offset(self, i, j):
        """Return the delay corresponding to baseline i,j."""
        return self.offs[self.bl_order[self.ij2bl(i,j)]]
    def gen_uvw(self, i, j, src='z'):
        """Compute uvw coordinates of baseline relative to provided RadioBody, 
        or 'z' for zenith uvw coordinates."""
        x,y,z = self.get_baseline(i,j, src=src)
        afreqs = self.ants[0].beam.afreqs
        if len(x.shape) == 0: return n.array([x*afreqs, y*afreqs, z*afreqs])
        afreqs = n.reshape(afreqs, (1,afreqs.size))
        x.shape += (1,); y.shape += (1,); z.shape += (1,)
        return n.array([n.dot(x,afreqs), n.dot(y,afreqs), n.dot(z,afreqs)])
    def gen_phs(self, src, i, j):
        """Return phasing that is multiplied to data to point to src."""
        try: src = src.e_vec
        except(AttributeError): pass
        bl = self.get_baseline(i,j,src='e')
        bl_dot_s = n.dot(src.transpose(), bl)
        if len(bl_dot_s.shape) >= 1: bl_dot_s.shape = (bl_dot_s.size, 1)
        t = self.get_delay(i,j)
        o = self.get_offset(i,j)
        afreqs = self.ants[0].beam.afreqs
        afreqs = n.reshape(afreqs, (1,afreqs.size))
        phs = n.exp(-1j*2*n.pi*(n.dot(bl_dot_s + t, afreqs) + o))
        return phs.squeeze()
    def resolve_src(self, src, i, j, srcshape=(0,0,0)):
        """Adjust amplitudes to reflect resolution effects for a uniform 
        elliptical disk characterized by srcshape:
        srcshape = (a1,a2,th) where a1,a2 are angular sizes along the 
            semimajor, semiminor axes, and th is the angle (in radians) of
            the semimajor axis from E."""
        u,v,w = self.gen_uvw(i, j, src=src)
        a1,a2,th = srcshape
        if len(u.shape) > len(a1.shape): 
            a1.shape += (1,); a2.shape += (1,); th.shape += (1,)
        ru = a1 * (u*n.cos(th) - v*n.sin(th))
        rv = a2 * (u*n.sin(th) + v*n.cos(th))
        x = 2 * n.pi * n.sqrt(ru**2 + rv**2)
        # Use first Bessel function of the first kind (J_1)
        return n.where(x == 0, 1, 2 * _cephes.j1(x)/x).squeeze()
    def phs2src(self, data, src, i, j):
        """Apply phasing to zenith-phased data to point to src."""
        return data * self.gen_phs(src, i, j)
    def unphs2src(self, data, src, i, j):
        """Remove phasing from src-phased data to point to zenith."""
        return data * n.conjugate(self.gen_phs(src, i, j))



#! /usr/bin/env python
"""
Apply the bandpass function in a UV file to the raw data, and then write
to a new UV file which will not have a bandpass file.  Has the (recommended)
option of linearizing for quantization gain effects.  For now, these are
applied only to the auto-correlations.  Cross-correlation quantization gain is
less sensitive to level, so linearization will come later.
Author: Aaron Parsons
Date: 8/14/07
Revisions:
    9/25/07 arp Uninverted bandpass and untransposed to sync with what
                Miriad expects from the bandpass variable.  Removes variables
                associated with bandpass (correctly).
    12/11/07 arp    Updated to new miriad file interface
    05/12/08 arp    Sped up reading w/ "raw" mode
"""

__version__ = '0.0.1'

import aipy as a, numpy as n, sys, os, optparse

o = optparse.OptionParser()
o.set_usage('apply_bp.py [options] *.uv')
o.set_description(__doc__)
o.add_option('-l', '--linearization', dest='linearization', default='comb',
    help='Apply the specified quantization linearization function to raw correlator values before applying bandpass.  Options are null, digi, full, and comb.  Default is comb')
o.add_option('-s', '--scale', dest='scale', type='float', default=12250000.,
    help='An additional numerical scaling to apply to the data.  Default: 12250000.')
opts, args = o.parse_args(sys.argv[1:])

# Digital Gain, Power Output (Channel 1024+512)
# NOD-5109 Noise Generator (Coarse=20dB, Fine=0dB attenuation)
# -3 dB, SHP-100, -3 dB, VLF-225
# port A pocket_correlator, shift=x3ff, clk=600MHz
# Data is [digital gain, power output]

qdata1 = n.array([
     [12157.543882402342, 69.423910140999993],
     [10941.789494162109, 66.790805816700001],
     [9847.6105447458976, 63.422136306799999],
     [8862.8494902713082, 59.848737716700001],
     [7976.5645412441772, 56.260694503800003],
     [7178.9080871197593, 52.3417243958],
     [6461.0172784077831, 48.231815338099999],
     [5814.9155505670051, 43.995954513500003],
     [5233.4239955103048, 39.623107910199998],
     [4710.0815959592746, 35.128669738799999],
     [4239.0734363633474, 30.719709396399999],
     [3815.1660927270127, 26.564506530799999],
     [3433.6494834543114, 22.680536270099999],
     [3090.2845351088804, 18.9764328003],
     [2781.2560815979923, 15.7225093842],
     [2503.1304734381933, 12.9534130096],
     [2252.8174260943742, 10.588439941400001],
     [2027.535683484937, 8.6283283233599999],
     [1824.7821151364433, 7.0549869537400003],
     [1642.303903622799, 5.7384414672900004],
     [1478.0735132605191, 4.6558799743700003],
     [1330.2661619344672, 3.8215141296400001],
     [1197.2395457410205, 3.1258945465100001],
     [1077.5155911669185, 2.5557441711400002],
     [969.76403205022666, 2.1025066375699999],
     [872.78762884520404, 1.72937583923],
     [785.50886596068369, 1.4391326904299999],
     [706.95797936461531, 1.19428634644],
     [636.26218142815378, 0.99624633789100003],
     [572.63596328533845, 0.83918571472199999],
     [515.37236695680463, 0.70255088806199995],
])

# Analog Gain (Coarse, Fine Attenuation), Power Output (Channel 1024+512)
# D-5109 Noise Generator (Coarse, Fine attenuation)
# -3 dB, SHP-100, -3 dB, VLF-225
# port A pocket_correlator, shift=x3ff, gain=[3500], clk=600MHz
# Data is [Attenuation (Coarse, Fine), Power Output]

qdata2 = n.array([
    [40,  4,  0.43  ],
    [40,  3,  0.46  ],
    [40,  2,  0.50  ],
    [40,  1,  0.55  ],
    [40,  0,  0.60  ],
    [30, 10,  0.60  ],
    [30,  9,  0.66  ],
    [30,  8,  0.74  ],
    [30,  7,  0.84  ],
    [30,  6,  1.0   ],
    [30,  5,  1.1   ],
    [30,  4,  1.3   ],
    [30,  3,  1.6   ],
    [30,  2,  1.9   ],
    [30,  1,  2.3   ],
    [30,  0,  2.7   ],
    [20, 10,  2.7   ],
    [20,  9,  3.4   ],
    [20,  8,  4.2   ],
    [20,  7,  5.2   ],
    [20,  6,  6.5   ],
    [20,  5,  8.1   ],
    [20,  4, 10.1   ],
    [20,  3, 12.6   ],
    [20,  2, 15.7   ],
    [20,  1, 19.3   ],
    [20,  0, 23.3   ],
    [10, 10, 23.7   ],
    [10,  9, 28.3   ],
    [10,  8, 32.9   ],
    [10,  7, 37.9   ],
    [10,  6, 42.6   ],
    [10,  5, 47.4   ],
    [10,  4, 51.9   ],
    [10,  3, 56.2   ],
    [10,  2, 60.2   ],
    [10,  1, 63.8   ],
    [10,  0, 66.9   ],
])

# Define some polynomials for interpreting between the datapoints above
# and use them to apply the correction function.  Polynomials are normalized
# to map a raw value of 10 to itself.

digi_cpoly = n.polyfit(qdata1[:,1], qdata1[:,0]**2, 6)
full_cpoly = n.polyfit(qdata2[:,2], 10**(-(qdata2[:,0]+qdata2[:,1])/10), 6)
digi_cpoly *= 10 / n.polyval(digi_cpoly, 10)
full_cpoly *= 10 / n.polyval(full_cpoly, 10)
null_cpoly = n.array([0, 0, 0, 0, 0, 1, 0])
# Unscientifically average the 3 linearization polynomials.  In practice, this
# seems to work best.
comb_cpoly = (null_cpoly + full_cpoly + digi_cpoly) / 3
cpolys = {'null':null_cpoly, 'digi':digi_cpoly, 
    'full':full_cpoly, 'comb':comb_cpoly}

cpoly = cpolys[opts.linearization]

print 'Using %s quantization correction' % opts.linearization

# These are all the items which should be removed once bandpass applied.
ignore_vars = ['bandpass', 'freqs', 'ngains', 'nspect0', 
    'nchan0', 'ntau', 'nfeeds', 'nsols', 'header', 'vartable']

# Process all files passed from the command line.
for filename in args:
    print filename,
    if os.path.exists(filename+'b'):
        print 'File exists: skipping'
        continue
    uvi = a.miriad.UV(filename)
    uvo = a.miriad.UV(filename+'b', status='new')
    uvo.init_from_uv(uvi, exclude=ignore_vars)
    nchan = uvi['nchan']
    nants = uvi['nants']
    try:
        # If there is a bandpass item, we're going apply it to data
        bp = uvi['bandpass'].real   # Sync'd with pocket_corr.py in corr pkg.
        bp.shape = (nants, nchan)
        print
    except:
        print 'No bandpass found'
        bp = n.ones((nants, nchan))
        print '.'
    def f(uv, preamble, data, flags):
        uvw, t, (i,j) = preamble
        if i == j: data = n.polyval(cpoly, data)
        data *= bp[i,:] * bp[j,:] * opts.scale
        return preamble, data, flags
    uvo.pipe(uvi, mfunc=f, raw=True,
        append2hist='APPLY_BP: ver=%s, corr type=%s, scale=%f\n' % \
            (__version__, opts.linearization, opts.scale))



#! /usr/bin/env python
"""
Print headers/variables in a Miriad UV file.  If no keywords are provided,
will print a list of available keywords.

Author: Aaron Parsons
"""

import aipy as a, sys, optparse

o = optparse.OptionParser()
o.set_usage('uvlist.py [options] *.uv')
o.set_description(__doc__)
o.add_option('-k', '--key', dest='key',
    help='Comma-delimited list of keywords to print for each UV file.')
opts,args = o.parse_args(sys.argv[1:])

for uvfile in args:
    print uvfile
    uv = a.miriad.UV(uvfile)
    if opts.key is None: print '    ', uv.items() + uv.vars()
    else:
        for key in opts.key.split(','):
            print '    ', key
            print '        ', uv[key]
    print '-----------------------------------------------------------'



#! /usr/bin/env python
"""
This is a general-purpose script for plotting simple FITS images.

Author: Aaron Parsons
"""

import aipy as a, numpy as n, pylab as p, sys, optparse, ephem, math
from mpl_toolkits.basemap import Basemap

o = optparse.OptionParser()
o.set_usage('plot_img.py [options] *.fits')
o.set_description(__doc__)
a.scripting.add_standard_options(o, chan=True)
o.add_option('-m', '--mode', dest='mode', default='log',
    help='Plot mode can be log (logrithmic), lin (linear), phs (phase), real, or imag.')
o.add_option('-o', '--out_file', dest='out_file', default='',
    help='If provided, will save the figure to the specified file instead of popping up a window.')
o.add_option('-p', '--pol', dest='pol', type='int', default=0, 
    help='Polarization index if FITS file has multiple polarizations.  Default 0.')
o.add_option('--max', dest='max', default=None, type='float',
    help='Upper clip value on 2D plots.')
o.add_option('--dyn_rng', dest='dyn_rng', default=None, type='float',
    help='Dynamic range in scale of 2D plots.')
opts, args = o.parse_args(sys.argv[1:])

m2 = int(math.sqrt(len(args)))
m1 = int(math.ceil(float(len(args)) / m2))

for cnt, filename in enumerate(args):
    # Gather data
    d, kwds = a.img.from_fits(filename)
    print kwds

    # Parse command-line options
    compress_axes = []
    ra_ax,dec_ax = (0,1)
    try:
        for i,ax in enumerate(kwds['axes']):
            if ax.startswith('ra'): ra_ax = i
            elif ax.startswith('dec'): dec_ax = i
            elif ax.startswith('freq'):
                chans = a.scripting.parse_chans(opts.chan, d.shape[i])
                d = d.take(chans, axis=i)
                compress_axes.append(i)
            elif ax.startswith('stokes'):
                d = d.take([opts.pol], axis=i)
                compress_axes.append(i)
    except(KeyError): pass

    compress_axes.reverse()
    for ax in compress_axes: d = n.average(d, axis=ax)

    # Put array in (ra,dec) order for plotting
    d = d.transpose((ra_ax,dec_ax))

    # Generate plots
    if opts.mode.startswith('phs'): d = n.angle(d.filled(0))
    elif opts.mode.startswith('lin'): d = n.ma.absolute(d)
    elif opts.mode.startswith('real'): d = d.real
    elif opts.mode.startswith('imag'): d = d.imag
    elif opts.mode.startswith('log'):
        d = n.ma.absolute(d)
        d = n.ma.masked_less_equal(d, 0)
        d = n.ma.log10(d)

    if not opts.max is None: max = opts.max
    else: max = d.max()
    if not opts.dyn_rng is None: min = max - opts.dyn_rng
    else: min = d.min()

    p.subplot(m2, m1, cnt+1)
    xpx,ypx = d.shape
    dx1 = -(xpx/2 + .5) * kwds['d_ra'] * a.img.deg2rad
    dx2 = (xpx/2 - .5) * kwds['d_ra'] * a.img.deg2rad
    dy1 = -(ypx/2 + .5) * kwds['d_dec'] * a.img.deg2rad
    dy2 = (ypx/2 - .5) * kwds['d_dec'] * a.img.deg2rad
    map = Basemap(projection='ortho', lon_0=180, lat_0=kwds['dec'],
        rsphere=1, llcrnrx=dx1, llcrnry=dy1, urcrnrx=dx2,urcrnry=dy2)
    map.drawmeridians(n.arange(kwds['ra']-180,kwds['ra']+180,30))
    map.drawparallels(n.arange(-90,120,30))
    map.drawmapboundary()
    map.imshow(d, vmin=min, vmax=max)
    p.colorbar(shrink=.5, fraction=.05)
    p.title(filename)

# Add right-click functionality for finding locations/strengths in map.
cnt = 1
def click(event):
    global cnt
    if event.button != 3: return
    lon,lat = map(event.xdata, event.ydata, inverse=True)
    lon = (180 + kwds['ra'] - lon) % 360
    lon *= a.img.deg2rad; lat *= a.img.deg2rad
    ra,dec = ephem.hours(lon), ephem.degrees(lat)
    xpx = n.around(event.xdata / (kwds['d_ra'] * a.img.deg2rad))
    ypx = n.around(event.ydata / (kwds['d_dec'] * a.img.deg2rad))
    flx = d[ypx,xpx]
    if opts.mode.startswith('log'): flx = 10**flx
    print '#%d (RA,DEC): (%s, %s), PX: (%d,%d) Jy: %f' % \
        (cnt, ra, dec, xpx, ypx, flx)
    cnt += 1

#register this function with the event handler
p.connect('button_press_event', click)

p.show()



"""
Module for gridding UVW data (including W projection), forming images,
and combining (mosaicing) images into spherical maps.
"""

import numpy as n, utils, coord, pyfits, time

deg2rad = n.pi / 180.
rad2deg = 180. / n.pi

def recenter(a, c):
    """Slide the (0,0) point of matrix a to a new location tuple c.  This is
    useful for making an image centered on your screen after performing an
    inverse fft of uv data."""
    s = a.shape
    c = (c[0] % s[0], c[1] % s[1])
    if n.ma.isMA(a):
        a1 = n.ma.concatenate([a[c[0]:], a[:c[0]]], axis=0)
        a2 = n.ma.concatenate([a1[:,c[1]:], a1[:,:c[1]]], axis=1)
    else:
        a1 = n.concatenate([a[c[0]:], a[:c[0]]], axis=0)
        a2 = n.concatenate([a1[:,c[1]:], a1[:,:c[1]]], axis=1)
    return a2

def convolve2d(a, b):
    """Convolve a and b by multiplying in Fourier domain.  Must be same size."""
    return n.fft.ifft2(n.fft.fft2(a) * n.fft.fft2(b))

def gaussian_beam(sigma, shape=0, amp=1., center=(0,0)):
    """Return a 2D gaussian.  Normalized to area under curve = 'amp'.  
    Down by 1/e at distance 'sigma' from 'center'."""
    if type(shape) == type(0): shape = array([2, 2]) * sigma
    def gaussian(x, y):
        nx = n.where(x > shape[0] / 2, x - shape[0], x)
        ny = n.where(y > shape[1] / 2, y - shape[1], y)
        return n.exp(-(nx**2 + ny**2) / sigma**2)
    g = n.fromfunction(gaussian, shape)
    g /= g.sum() / amp
    return recenter(g, center)

def beam_gain(bm):
    return n.sqrt((n.abs(bm)**2).sum())

class Img:
    """Class for gridding uv data, recording the synthesized beam profile,
    and performing transforms into image domain."""
    def __init__(self, size=100, res=1, mf_order=0):
        """size = number of wavelengths which the UV matrix spans (this 
        determines the image resolution).
        res = resolution of the UV matrix (determines image field of view)."""
        self.res = float(res)
        self.size = float(size)
        dim = n.round(self.size / self.res)
        self.shape = (dim,dim)
        self.uv = n.zeros(shape=self.shape, dtype=n.complex64)
        self.bm = []
        for i in range(mf_order+1):
            self.bm.append(n.zeros(shape=self.shape, dtype=n.complex64))
    def get_LM(self, center=(0,0)):
        """Get the (l,m) image coordinates for an inverted UV matrix."""
        dim = self.shape[0]
        M,L = n.indices(self.shape)
        L,M = n.where(L > dim/2, dim-L, -L), n.where(M > dim/2, M-dim, M)
        L,M = L.astype(n.float)/dim/self.res, M.astype(n.float)/dim/self.res
        mask = n.where(L**2 + M**2 >= 1, 1, 0)
        L,M = n.ma.array(L, mask=mask), n.ma.array(M, mask=mask)
        return recenter(L, center), recenter(M, center)
    def put(self, (u,v,w), data, wgts=None, apply=True):
        """Grid uv data (w is ignored) onto a UV plane.  Data should already
        have the phase due to w removed.  Assumes the Hermitian conjugate
        data is in uvw already (i.e. the conjugate points are not placed for
        you).  If wgts are not supplied, default is 1 (normal weighting).
        If apply is false, returns uv and bm data without applying it do
        the internally stored matrices."""
        if wgts is None:
            wgts = []
            for i in range(len(self.bm)):
                if i == 0: wgts.append(n.ones_like(data))
                else: wgts.append(n.zeros_like(data))
        if len(self.bm) == 1 and len(wgts) != 1: wgts = [wgts]
        assert(len(wgts) == len(self.bm))
        if apply: uv,bm = self.uv,self.bm
        else:
            uv = n.zeros_like(self.uv)
            bm = [n.zeros_like(i) for i in self.bm]
        u = n.round(u / self.res).astype(n.int)
        v = n.round(v / self.res).astype(n.int)
        inds = n.array([-v,u],).transpose()
        ok = n.logical_and(n.abs(inds[:,0]) < self.shape[0],
            n.abs(inds[:,1]) < self.shape[1])
        data = data.compress(ok)
        inds = inds.compress(ok, axis=0)
        utils.add2array(uv, inds, data.astype(uv.dtype))
        for i,wgt in enumerate(wgts):
            wgt = wgt.compress(ok)
            utils.add2array(bm[i], inds, wgt.astype(bm[0].dtype))
        if not apply: return uv, bm
    def append_hermitian(self, (u,v,w), data, wgts=None):
        """Append to (uvw, data, [wgts]) the points (-uvw, conj(data), [wgts]).
        This is standard practice to get a real-valued image."""
        u = n.concatenate([u, -u], axis=0)
        v = n.concatenate([v, -v], axis=0)
        w = n.concatenate([w, -w], axis=0)
        data = n.concatenate([data, n.conj(data)], axis=0)
        if wgts is None: return n.array((u,v,w)), data
        if len(self.bm) == 1 and len(wgts) != 1: wgts = [wgts]
        assert(len(wgts) == len(self.bm))
        for i,wgt in enumerate(wgts): wgts[i] = n.concatenate([wgt,wgt],axis=0)
        return (u,v,w), data, wgts
    def _gen_img(self, data, center=(0,0)):
        """Return the inverse FFT of the provided data, with the 0,0 point 
        moved to 'center'.  Up=North, Right=East."""
        return recenter(n.abs(n.fft.ifft2(data)), center)
    def image(self, center=(0,0)):
        """Return the inverse FFT of the UV matrix, with the 0,0 point moved
        to 'center'.  Tranposes to put up=North, right=East."""
        return self._gen_img(self.uv, center=center)
    def bm_image(self, center=(0,0), term=None):
        """Return the inverse FFT of the sample weightings (for all mf_order
        terms, or the specified term if supplied), with the 0,0 point
        moved to 'center'.  Tranposes to put up=North, right=East."""
        if not term is None:
            return self._gen_img(self.bm[term], center=center)
        else:
            return [self._gen_img(b, center=center) for b in self.bm]
    def uniform_wgt(self, thresh=.1):
        """Reweight data in the UV/BM to reflect uniform weighting above
        the specified threshold (expressed as a fraction of the maximum
        weight in the beam."""
        wgts = n.abs(self.bm[0])
        thresh = wgts.max() * thresh
        divisor = wgts.clip(thresh, n.Inf)
        self.uv /= divisor
        for i in range(len(self.bm)): self.bm[i] /= divisor
    def get_top(self, center=(0,0)):
        """Return the topocentric coordinates of each pixel in the image."""
        x,y = self.get_LM(center)
        z = n.sqrt(1 - x**2 - y**2)
        return x,y,z
    def get_eq(self, ra=0, dec=0, center=(0,0)):
        """Return the equatorial coordinates of each pixel in the image, 
        assuming the image is centered on the provided ra, dec (in radians)."""
        x,y,z = self.get_top(center)
        shape,mask = x.shape, x.mask
        vec = n.array([a.filled().flatten() for a in (x,y,z)])
        m = coord.top2eq_m(-ra, dec)
        vec = n.dot(m, vec)
        vec.shape = (3,) + shape
        return n.ma.array(vec, mask=[mask,mask,mask])

class ImgW(Img):
    """A subclass of Img adding W projection functionality (see Cornwell
    et al. 2005 "Widefield Imaging Problems in Radio Astronomy")."""
    def __init__(self, size=100, res=1, wres=.5, mf_order=0):
        """wres: the gridding resolution of sqrt(w) when projecting to w=0."""
        Img.__init__(self, size=size, res=res, mf_order=mf_order)
        self.wres = wres
    def put(self, (u,v,w), data, wgts=None, invker2=None):
        """Same as Img.put, only now the w component is projected to the w=0
        plane before applying the data to the UV matrix."""
        if len(u) == 0: return
        if wgts is None:
            wgts = []
            for i in range(len(self.bm)):
                if i == 0: wgts.append(n.ones_like(data))
                else: wgts.append(n.zeros_like(data))
        if len(self.bm) == 1 and len(wgts) != 1: wgts = [wgts]
        assert(len(wgts) == len(self.bm))
        # Sort uvw in order of w
        order = n.argsort(w)
        u = u.take(order)
        v = v.take(order)
        w = w.take(order)
        data = data.take(order)
        wgts = [wgt.take(order) for wgt in wgts]
        sqrt_w = n.sqrt(n.abs(w)) * n.sign(w)
        i = 0
        while True:
            # Grab a chunk of uvw's that grid w to same point.
            j = sqrt_w.searchsorted(sqrt_w[i]+self.wres)
            avg_w = n.average(w[i:j])
            # Put all uv's down on plane for this gridded w point
            wgtsij = [wgt[i:j] for wgt in wgts]
            uv,bm = Img.put(self, (u[i:j],v[i:j],w[i:j]),
                data[i:j], wgtsij, apply=False)
            # Convolve with the W projection kernel
            invker = n.fromfunction(lambda u,v: self.conv_invker(u,v,avg_w),
                uv.shape)
            if not invker2 is None: invker *= invker2
            self.uv += n.fft.ifft2(n.fft.fft2(uv) * invker)
            for b in range(len(self.bm)):
                self.bm[b] += n.fft.ifft2(n.fft.fft2(bm[b]) * invker)
            if j >= len(w): break
            i = j
    def conv_invker(self, u, v, w):
        """Generates the W projection kernel (a function of u,v) for the
        supplied value of w.  See Cornwell et al. 2005 "Widefield Imaging
        Problems in Radio Astronomy" for discussion.  This implementation
        uses a numerically evaluated Fresnel kernel, rather than the
        small-angle approximated one given in the literature."""
        L,M = self.get_LM()
        # This is the exactly evaluated kernel (works better)
        sqrt = n.sqrt(1 - L**2 - M**2)
        G = n.exp(-2*n.pi*1j*w*(sqrt - 1))
        # This is the kernel described by Cornwell using the small angle approx.
        #G = n.exp(n.pi*1j*w*(l**2 + m**2))
        G = G.filled(0)
        # Unscramble difference between fft(fft(G)) and G
        G[1:] = n.flipud(G[1:]).copy()
        G[:,1:] = n.fliplr(G[:,1:]).copy()
        return G / G.size

'''
class ImgBW(ImgW):
    """A subclass of ImgW adding better weighting for the primary beam."""
    def __init__(self, size=100, res=1, wres=.5, tres=.2):
        """wres: the gridding resolution of sqrt(w) when projecting to w=0."""
        ImgW.__init__(self, size=size, res=res, wres=wres)
        self.tres = tres
    def put(self, uvw, data, wgts, has, odec, sdec, bm_response):
        """Same as Img.put, only now the w component is projected to the w=0
        plane before applying the data to the UV matrix."""
        if len(uvw) == 0: return
        order = n.argsort(has)
        # Sort uvw in order of time
        uvw = uvw.take(order, axis=0)
        data = data.take(order)
        wgts = wgts.take(order)
        i = 0
        while True:
            # Grab a chunk of uvw's that have same time
            j = has.searchsorted(has[i]+self.tres)
            ha = n.average(has[i:j])
            eqs = self.get_eq(ra=-ha, dec=sdec)
            mask = eqs[0].mask; eqs = eqs.filled(0)
            eqs.shape = (3, eqs.size / 3)
            m = coord.eq2top_m(0, odec)
            top = n.dot(m, eqs)
            z = top[-1]; z.shape = self.uv.shape
            mask = n.logical_or(mask, z < 0)
            resp = bm_response(top); resp.shape = self.uv.shape
            resp = n.where(mask, 0, resp)
            if resp[0,0] <= 0:
                if j >= len(has): break
                i = j
                continue
            resp = resp[0,0] / resp.clip(min(resp[0,0], resp.max()/4),n.Inf)
            #resp = n.where(mask, 0, resp)
            # Transpose here to match transposition in image() and bm_image()
            resp = resp.transpose()
            # Unscramble difference between fft(fft(resp)) and resp
            resp[1:] = n.flipud(resp[1:]).copy()
            resp[:,1:] = n.fliplr(resp[:,1:]).copy()
            resp /= resp.size
            # Put all uv's down on plane for this time
            ImgW.put(self, uvw[i:j,:],data[i:j],wgts[i:j],invker2=resp)
            if j >= len(has): break
            i = j
'''
default_fits_format_codes = {
    n.bool_:'L', n.uint8:'B', n.int16:'I', n.int32:'J', n.int64:'K',
    n.float32:'E', n.float64:'D', n.complex64:'C', n.complex128:'M'
}

def to_fits(filename, data, clobber=False,
        axes=('ra---sin','dec---sin','freq','stokes'),
        object='', telescope='', instrument='', observer='', origin='AIPY',
        obs_date=time.strftime('%D'), cur_date=time.strftime('%D'), 
        ra=0, dec=0, d_ra=0, d_dec=0, epoch=2000., 
        freq=0, d_freq=0, bscale=0, bzero=0):
    """Write image data to a FITS file.  Follows convention of VLA image
    headers.  "axes" describes dimensions of "data" provided.  (ra,dec) are
    the degree coordinates of image center in the specified "epoch". 
    (d_ra,d_dec) are approximate pixel-deltas for ra,dec (approximate because 
    if sine projections of these coordinates are specified--e.g. 
    "ra---sin"--then the deltas change away from the image center).  If a 
    "freq" axis is specified, then "freq" is the frequency of the first entry 
    (in Hz), and "d_freq" is the width of the channel.  The rest are pretty 
    self-explanitory/can be used however you want."""
    data.shape = data.shape + (1,) * (len(axes) - len(data.shape))
    phdu = pyfits.PrimaryHDU(data)
    phdu.data = data.transpose()
    phdu.update_header()
    phdu.header.update('OBJECT', object, comment='SOURCE NAME')
    phdu.header.update('TELESCOP', telescope)
    phdu.header.update('INSTRUME', instrument)
    phdu.header.update('OBSERVER', observer)
    phdu.header.update('DATE-OBS', obs_date, 
        comment='OBSERVATION START DATE DD/MM/YY')
    #phdu.header.update('DATE-MAP', '', 
    #    comment='DATE OF LAST PROCESSING DD/MM/YY')
    phdu.header.update('BSCALE ', bscale,
        comment='REAL = FITS_VALUE * BSCALE + BZERO')
    phdu.header.update('BZERO  ', bzero)
    phdu.header.update('BUNIT  ', 'JY/BEAM ', comment='UNITS OF FLUX')
    phdu.header.update('EQUINOX', epoch, comment='EQUINOX OF RA DEC')
    phdu.header.update('DATAMAX', data.max(), comment='MAX PIXEL VALUE')
    phdu.header.update('DATAMIN', data.min(), comment='MIN PIXEL VALUE')
    for i,ax in enumerate(axes):
        if ax.lower().startswith('ra'): val,delta = (ra, d_ra)
        elif ax.lower().startswith('dec'): val,delta = (dec, d_dec)
        elif ax.lower().startswith('freq'): val,delta = (freq, d_freq)
        elif ax.lower().startswith('stokes'): val,delta = (1, 1)
        else: val,delta = (0,0)
        phdu.header.update('CTYPE%d' % (i+1), ax.upper())
        phdu.header.update('CRVAL%d' % (i+1), val)
        phdu.header.update('CDELT%d' % (i+1), delta)
        phdu.header.update('CROTA%d' % (i+1), 0)
        phdu.header.update('CRPIX%d' % (i+1), phdu.data.shape[-(i+1)])
    phdu.header.update('ORIGIN', origin)
    phdu.header.update('DATE', cur_date, comment='FILE WRITTEN ON DD/MM/YY')
    pyfits.writeto(filename, phdu.data, phdu.header, clobber=True)

def from_fits(filename):
    """Read (data,kwds) from a FITS file.  Matches to_fits() above.  Attempts
    to deduce each keyword listed in to_fits() from the FITS header, but is
    accepting of differences.  Returns values in "kwds" dictionary."""
    phdu = pyfits.open(filename)[0]
    data = phdu.data.transpose()
    kwds = {}
    hitems = (('OBJECT','object'), ('TELESCOP','telescope'),
        ('INSTRUME','instrument'), ('OBSERVER','observer'),
        ('DATE-OBS','obs_date'), ('BSCALE','bscale'), ('BZERO','bzero'),
        ('EQUINOX','epoch'), ('ORIGIN','origin'), ('DATE','cur_date'))
    for fitsname,name in hitems:
        try: kwds[name] = phdu.header[fitsname]
        except(KeyError): pass
    axes = []
    for i in range(phdu.header['NAXIS']):
        try:
            ax = phdu.header['CTYPE%d' % (i+1)].lower()
            axes.append(ax)
            val = phdu.header['CRVAL%d' % (i+1)]
            delta = phdu.header['CDELT%d' % (i+1)]
            if ax.startswith('ra'): kwds['ra'],kwds['d_ra'] = (val,delta)
            elif ax.startswith('dec'): kwds['dec'],kwds['d_dec'] = (val,delta)
            elif ax.startswith('freq'): kwds['freq'],kwds['d_freq']=(val,delta)
            elif ax.startswith('stokes'): pass
            else: pass
        except(KeyError): pass
    kwds['axes'] = axes
    return data, kwds



#
# optimize - Optimization Tools
#

from info import __doc__

from optimize import *
from anneal import *
from nonlin import broyden1, broyden2, broyden3, broyden_generalized, \
    anderson, anderson2 

__all__ = filter(lambda s:not s.startswith('_'),dir())



#! /usr/bin/env python
import aipy.healpix, sys, numpy

alm = aipy.healpix.Alm(1,1)
alm = aipy.healpix.Alm(2,1)
try:
    print "Checking that exception gets thrown:"
    alm = aipy.healpix.Alm(1,2)
    print "Should have failed (but didn't)"
    sys.exit(0)
except(RuntimeError): print "Ok."
alm = aipy.healpix.Alm(2,2)
print 'Lmax:', alm.lmax(), 'Mmax:', alm.mmax()
data = alm.get_data()
data[2] = 100
try: alm.set_data(data[:-1])
except(ValueError): pass
alm.set_data(data)
print alm.get_data(data)
for l,m in zip(*alm.lm_indices()):
    print l,m, '->', alm[l,m], '->',
    alm[l,m] = l+m*1j
    print alm[l,m]
print alm.get_data()
alm = aipy.healpix.Alm(2,2)
alm[1,0] = 1
alm[1,0] = 1.
alm[1,0] = 1j
try: alm[1,0] = "hi" ; print alm[1,0]
except(ValueError): pass
alm = aipy.healpix.Alm(2,2)
alm[0,0] = 4
hmap = aipy.healpix.HealpixMap(64, scheme="RING")
hmap.from_alm(alm)
#hmap.SetData(alm.to_map(hmap.Nside(), hmap.Scheme()))
print "Actual coeffiecients:"
print alm.get_data()
alm = hmap.to_alm(2, 2) ; print alm.get_data()
#alm.from_map(hmap.map.astype(numpy.int), 1); print alm.get_data()
print "long test:"
alm.from_map(hmap.map.astype(numpy.long), 1); print alm.get_data()
print "float test:"
alm.from_map(hmap.map.astype(numpy.float32), 1); print alm.get_data()
print "double test:"
for i in [1,2,3]:
    print "iter %d:" % i
    alm.from_map(hmap.map.astype(numpy.double), i); print alm.get_data()
#hmap.to_fits('alm_test.fits')



#! /usr/bin/env python
"""
Creates waterfall plots from Miriad UV files.  Can tile multiple plots
on one window, or plot just a single baseline.  When taking the delay
transform (-d), channels are interepreted as selecting delays after the
the transform operation.  Similarly, the time select (-t) will be interpreted
as selecting fringe rates if the fringe-rate transform (-f) has been selected.
In both cases, the ranges specified are intepreted to be the units of the
output plot (i.e. as specified by --chan_axis and --time_axis).

Author: Aaron Parsons, Griffin Foster
"""

import aipy as a, numpy as n, pylab as p, math, sys, optparse

o = optparse.OptionParser()
o.set_usage('plot_uv.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, ant=True, pol=True, chan=True, dec=True)
o.add_option('-m', '--mode', dest='mode', default='log',
    help='Plot mode can be log (logrithmic), lin (linear), phs (phase), real, or imag.')
o.add_option('--sum_chan', dest='sum_chan', action='store_true',
    help='Sum active channels together.')
o.add_option('-t', '--time', dest='time', default='all', help='Select which time sample to plot. Options are: "all" (default), "<time1 #>_<time2 #>" (a range of times to plot), or "<time1 #>,<time2 #>" (a list of times to plot). If "all" or a range are selected, a 2-d image will be plotted. If a list of times is selected an xy plot will be generated.')
o.add_option('-u', '--unmask', dest='unmask', action='store_true',
    help='Plot masked data, too.')
o.add_option('-d', '--delay', dest='delay', action='store_true',
    help='Take FFT of frequency axis to go to delay (t) space.')
o.add_option('-f', '--fringe', dest='fringe', action='store_true',
    help='Take FFT of time axis to go to fringe (Hz) space.')
o.add_option('--dt', dest='dt', action='store_true',
    help='Remove a linear extrapolation from adjacent times.')
o.add_option('--df', dest='df', action='store_true',
    help='Remove a linear extrapolation from adjacent frequency channels.')
o.add_option('-o', '--out_file', dest='out_file', default='',
    help='If provided, will save the figure to the specified file instead of popping up a window.')
o.add_option('--max', dest='max', default=None, type='float', 
    help='Upper clip value on 2D plots.')
o.add_option('--dyn_rng', dest='dyn_rng', default=None, type='float', 
    help='Dynamic range in scale of 2D plots.')
o.add_option('--time_axis', dest='time_axis', default='index',
    help='Choose time axis to be integration/fringe index (index), or physical coordinates (physical), or if doing xy plot in time-mode, (lst) is also available.  Default is index.')
o.add_option('--chan_axis', dest='chan_axis', default='index',
    help='Choose channel axis to be channel/delay index (index), or physical coordinates (physical).  Default is index.')
o.add_option('--clean', dest='clean', type='float',
    help='Deconvolve delay-domain data by the "beam response" that results from flagged data.  Specify a tolerance for termination (usually 1e-2 or 1e-3).')

def convert_arg_range(arg):
    """Split apart command-line lists/ranges into a list of numbers."""
    arg = arg.split(',')
    return [map(float, option.split('_')) for option in arg]

def gen_chans(chanopt, uv, coords, is_delay):
    """Return an array of active channels and whether or not a range of
    channels is selected (as opposed to one or more individual channels)
    based on command-line arguments."""
    is_chan_range = True
    if chanopt == 'all': chans = n.arange(uv['nchan'])
    else:
        chanopt = convert_arg_range(chanopt)
        if coords != 'index':
            if is_delay:
                def conv(c):
                    return int(n.round(c * uv['sdf'] * uv['nchan'])) \
                        + uv['nchan']/2
            else:
                def conv(c): return int(n.round((c - uv['sfreq']) / uv['sdf']))
        else:
            if is_delay:
                def conv(c): return int(c) + uv['nchan']/2
            else:
                def conv(c): return c
        chanopt = [map(conv, c) for c in chanopt]
        if len(chanopt[0]) != 1: 
            chanopt = [n.arange(x,y, dtype=n.int) for x,y in chanopt]
        else: is_chan_range = False
        chans = n.concatenate(chanopt)
    return chans.astype(n.int), is_chan_range

def gen_times(timeopt, uv, coords, decimate, is_fringe):
    is_time_range = True
    if timeopt == 'all' or is_fringe:
        def time_selector(t, cnt): return True
    else:
        timeopt = convert_arg_range(timeopt)
        if len(timeopt[0]) != 1:
            def time_selector(t, cnt):
                if coords == 'index': t = cnt
                for opt in timeopt:
                    if (t >= opt[0]) and (t < opt[1]): return True
                return False
        else:
            is_time_range = False
            timeopt = [opt[0] for opt in timeopt]
            inttime = uv['inttime'] / a.const.s_per_day * decimate
            def time_selector(t, cnt):
                if coords == 'index': return cnt in timeopt
                for opt in timeopt:
                    if (t >= opt) and (t < opt + inttime): return True
                return False
    return time_selector, is_time_range

opts, args = o.parse_args(sys.argv[1:])

# Parse command-line options
uv = a.miriad.UV(args[0])
a.scripting.uv_selector(uv, opts.ant, opts.pol)
chans, is_chan_range = gen_chans(opts.chan, uv, opts.chan_axis, opts.delay)
freqs = n.arange(uv['sfreq'], uv['sfreq']+uv['nchan']*uv['sdf'], uv['sdf'])
freqs = freqs.take(chans)
delays = n.arange(-.5/uv['sdf'], .5/uv['sdf'], 1/(uv['sdf']*uv['nchan']))
delays = delays.take(chans)
time_sel, is_time_range = gen_times(opts.time, uv, opts.time_axis, 
    opts.decimate, opts.fringe)
inttime = uv['inttime'] * opts.decimate
del(uv)

# Loop through UV files collecting relevant data
plot_x = {}
plot_t = {'jd':[], 'lst':[], 'cnt':[]}
times = []

for uvfile in args:
    print 'Reading', uvfile
    uv = a.miriad.UV(uvfile)
    # Only select data that is needed to plot
    a.scripting.uv_selector(uv, opts.ant, opts.pol)
    # Read data from a single UV file
    for (uvw,t,(i,j)),d in uv.all():
        bl = '%d,%d' % (i,j)
        # Implement Decimation
        if len(times) == 0 or times[-1] != t:
            times.append(t)
            use_this_time = ((len(times) - 1) % opts.decimate) == 0
            use_this_time &= time_sel(t, (len(times)-1) / opts.decimate)
            if use_this_time:
                plot_t['lst'].append(uv['lst'])
                plot_t['jd'].append(t)
                plot_t['cnt'].append((len(times)-1) / opts.decimate)
        if not use_this_time: continue
        # Do delay transform if required
        if opts.delay:
            if opts.unmask:
                d = d.data
                ker = n.zeros_like(d)
                ker[0] = 1.
                gain = 1.
            else:
                flags = n.logical_not(d.mask).astype(n.float)
                gain = n.sqrt(n.average(flags**2))
                ker = n.fft.ifft(flags)
                d = d.filled(0)
            d = n.fft.ifft(d)
            if not opts.clean is None and not n.all(d == 0):
                d, info = a.deconv.clean1d(d, ker, tol=opts.clean)
                d += info['res'] / gain
            d = n.ma.array(d)
            d = n.ma.concatenate([d[d.shape[0]/2:], d[:d.shape[0]/2]], 
                axis=0)
        elif opts.unmask: d = d.data
        # Extract specific channels for plotting
        d = d.take(chans)
        d.shape = (1,) + d.shape
        if not plot_x.has_key(bl): plot_x[bl] = []
        plot_x[bl].append(d)
    del(uv)

bls = plot_x.keys()
def sort_func(a, b):
    ai,aj = map(int, a.split(','))
    bi,bj = map(int, b.split(','))
    if bi > ai or (bi == ai and bj > aj): return -1
    return 1
bls.sort(cmp=sort_func)
if len(bls) == 0:
    print 'No data to plot.'
    sys.exit(0)
m2 = int(math.sqrt(len(bls)))
m1 = int(math.ceil(float(len(bls)) / m2))

# Generate all the plots
for cnt, bl in enumerate(bls):
    d = n.ma.concatenate(plot_x[bl], axis=0)
    if opts.df: d = d[:,:-2]/2 + d[:,2:]/2 - d[:,1:-1]
    if opts.dt: d = d[:-2]/2 + d[2:]/2 - d[1:-1]
    if opts.fringe:
        d = d.filled(0)
        flags = n.where(d[:,0] != 0, 1., 0.)
        gain = n.sqrt(n.average(flags**2))
        ker = n.fft.ifft(flags)
        d = n.fft.ifft(d, axis=0)
        if not opts.clean is None:
            for chan in range(d.shape[1]):
                d[:,chan],info = a.deconv.clean1d(d[:,chan],ker,tol=opts.clean)
                d[:,chan] += info['res'] / gain
        d = n.ma.concatenate([d[d.shape[0]/2:], d[:d.shape[0]/2]], axis=0)
    if opts.sum_chan:
        d = d.sum(axis=1)
        is_chan_range = False
    if opts.mode.startswith('phs'): d = n.angle(d.filled(0))
    elif opts.mode.startswith('lin'): d = n.ma.absolute(d)
    elif opts.mode.startswith('real'): d = d.real
    elif opts.mode.startswith('imag'): d = d.imag
    elif opts.mode.startswith('log'):
        d = n.ma.absolute(d.filled(0))
        d = n.ma.masked_less_equal(d, 0)
        d = n.ma.log10(d)
    else: raise ValueError('Unrecognized plot mode.')
    p.subplot(m2, m1, cnt+1)
    if is_chan_range and is_time_range:
        if opts.fringe:
            if opts.time_axis == 'index':
                if opts.time != 'all':
                    t1, t2 = map(float, opts.time.split('_'))
                    d = d[t1+d.shape[0]/2:t2+d.shape[0]/2]
                else:
                    t1 = len(plot_t['jd'])/2 - len(plot_t['jd'])
                    t2 = len(plot_t['jd'])/2
                ylabel = 'Fringe Rate (bins)'
            else:
                t1 = -500/inttime
                t2 =  500/inttime - 1000 / (inttime * len(plot_t['jd']))
                ylabel = 'Fringe Rate (milliHz)'
        else:
            if opts.time_axis == 'index':
                t1,t2 = plot_t['cnt'][0], plot_t['cnt'][-1]
                ylabel = 'Time (integrations)'
            else:
                t1,t2 = plot_t['jd'][0], plot_t['jd'][-1]
                ylabel = 'Time (Julian Date)'
        if opts.delay:
            if opts.chan_axis == 'index':
                c1,c2 = len(chans)/2 - len(chans), len(chans)/2
                xlabel = 'Delay (bins)'
            else:
                c1,c2 = delays[0], delays[-1]
                xlabel = 'Delay (ns)'
        else:
            if opts.chan_axis == 'index':
                c1,c2 = 0, len(chans) - 1
                xlabel = 'Frequency (chan)'
            else:
                c1,c2 = freqs[0], freqs[-1]
                xlabel = 'Frequency (GHz)'
        if not opts.max is None: max = opts.max
        else: max = d.max()
        if not opts.dyn_rng is None: min = max - opts.dyn_rng
        else: min = d.min()
        p.imshow(d, extent=(c1,c2,t2,t1), aspect='auto', vmax=max, vmin=min)
        p.colorbar()
        p.xlabel(xlabel); p.ylabel(ylabel)
    elif is_chan_range and not is_time_range:
        if opts.delay:
            if opts.chan_axis == 'index':
                plot_chans = range(len(chans)/2 - len(chans), len(chans)/2)
                xlabel = 'Delay (bins)'
            else:
                plot_chans = delays
                xlabel = 'Delay (ns)'
        else:
            if opts.chan_axis == 'index':
                plot_chans = chans
                xlabel = 'Frequency (chan)'
            else:
                plot_chans = freqs
                xlabel = 'Frequency (GHz)'
        if cnt == 0:
            if opts.time_axis == 'index':
                plot_t = plot_t['cnt']
                label = '#%d'
            else:
                plot_t = plot_t['jd']
                label = 'jd%f'
        for i,t in enumerate(plot_t):
            p.plot(plot_chans, d[i,:], '-', label=label % t)
        p.xlabel(xlabel)
        if not opts.max is None: max = opts.max
        else: max = d.max()
        if not opts.dyn_rng is None: min = max - opts.dyn_rng
        else: min = d.min()
        p.ylim(min,max)
    elif not is_chan_range and is_time_range:
        if opts.time_axis == 'index': plot_times = range(len(plot_t['jd']))
        elif opts.time_axis == 'physical': plot_times = plot_t['jd']
        elif opts.time_axis == 'lst': plot_times = plot_t['lst']
        else: raise ValueError('Unrecognized time axis type.')
        if opts.sum_chan: p.plot(plot_times, d, '.', label='(+)')
        else:
            if opts.chan_axis == 'index': label = '#%d'
            else:
                chans = freqs
                label = '%f GHz'
            for c, chan in enumerate(chans):
                p.plot(plot_times, d[:,c], '.', label=label % chan)
        if not opts.max is None: max = opts.max
        else: max = d.max()
        if not opts.dyn_rng is None: min = max - opts.dyn_rng
        else: min = d.min()
        p.ylim(min,max)
    else: raise ValueError('Either time or chan needs to be a range.')
    p.title(bl)
if not is_time_range or not is_chan_range: p.legend(loc='best')

# Save to a file or pop up a window
if opts.out_file != '': p.savefig(opts.out_file)
else: p.show()

